
class ColumnNameChangeWarning(UserWarning):
    """Warning that causallib renamed input name
    to ensure all columns are of a single type
    so scikit-learn>=1.2.0 is happy.

    See array validation:
    https://github.com/scikit-learn/scikit-learn/blob/8133ecaacca77f06a8c4c560f5dbbfd654f1990f/sklearn/utils/validation.py#L2271-L2280"""
    pass



from . import BasePositivity
import pandas as pd
import numpy as np
from sklearn.exceptions import NotFittedError


class Support:
    """Base Support class
    """

    def __init__(self, support=None):
        self.support = support

    def __str__(self):
        return "support: " + self.__repr__()

    def __repr__(self):
        if self.support is None:
            return "no support"
        return repr(self.support)

    def assert_same_type(self, other):
        if not isinstance(other, type(self)) and not isinstance(self, type(other)):
            raise ValueError(
                "Cannot intersect discrete and continuous support")


class ContinuousSupport(Support):
    """Continuous support class

    Intended for use with continuous valued data for which `min` and `max` can
    be said to characterize the support.
    """

    def fit(self, X):
        """Calculate support as min and max of continuous data `X`

        Args:
            X (pd.DataFrame|pd.Series): one dimensional continuous valued data

        Returns:
            ContinuousSupport: fitted ContinuousSupport object
        """
        self.support = [min(X), max(X)]
        return self

    def predict(self, x):
        """Predict if variable is in support

        Args:
            x (int|float): numerical value to check

        Returns:
            bool: True if `x` in support else False
        """
        if x >= self.support[0] and x <= self.support[1]:
            return True
        return False

    def intersection(self, other_support):
        """Find intersection of supports

        Args:
            other_support (ContinuousSupport): other support

        Returns:
            ContinuousSupport: intersection of this support with other support

        Raises:
            ValueError: if attempting to intersect with incompatible type
        """
        self.assert_same_type(other_support)
        if self.non_zero_overlap(other_support):
            joint_support = [max(self.support[0], other_support.support[0]),
                             min(self.support[1], other_support.support[1])]
        else:
            joint_support = None
        return ContinuousSupport(support=joint_support)

    def non_zero_overlap(self, other_support):
        if self.support is None or other_support.support is None:
            return False
        if self.support[1] >= other_support.support[0] and other_support.support[1] >= self.support[0]:
            return True

    def __sub__(self, other):
        return [self.support[0] - other.support[0], self.support[1] - other.support[1]]


class CategoricalSupport(Support):
    """Support for categorical variables based on `set`
    """

    def __init__(self, support=None):
        super().__init__(support=support)
        if self.support is not None and not isinstance(self.support, set):
            self.support = set(self.support)

    def fit(self, X):
        """Calculate support of categorical variables in `X` using `set`

        Args:
            X (pd.Series|pd.DataFrame): one dimensional categorical data

        Returns:
            CategoricalSupport: fitted discrete support object
        """
        self.support = set(X)
        return self

    def predict(self, x):
        return True if x in self.support else False

    def intersection(self, other_support):
        self.assert_same_type(other_support)
        return CategoricalSupport(support=self.support.intersection(other_support.support))

    def __sub__(self, other):
        return (self.support - other.support).union(other.support - self.support)


class QuantileContinuousSupport(ContinuousSupport):
    """Continuous support based on quantiles
    """

    def __init__(self, alpha=0.01, support=None):
        super().__init__(support=support)
        self.alpha = alpha

    def fit(self, X):
        """Calculate support based on quantiles

        Args:
            X (pd.DataFrame|pd.Series): one dimensional continuous valued data

        Returns:
            QuantileContinuousSupport: fitted quantile Continuous support object
        """
        self.support = list(np.quantile(
            X.values, [self.alpha/2, 1 - (self.alpha/2)]))
        return self


class UnivariateBoundingBox(BasePositivity):
    """Filter positivity by calculating univariate support
    """

    def __init__(self, quantile_alpha=0.1, continuous_columns=[], categorical_columns=[]):
        """

        Args:
            quantile_alpha (float, optional): Quantile cut-off for continuous
                variable support calculation. If not None, then the support for
                the continuous variables will be calculated using the data at 
                quantile quantile_alpha/2 as the left end and quantile 
                1 - quantile_alpha/2 on the right end. Defaults to 0.1.
            continuous_columns (List[str], optional): Column names to
                treat as Continuous variables. Defaults to None.
            categorical_columns (List[str], optional): Column names to
                treat as categorical variables. Defaults to None.

        """
        self.quantile_alpha = quantile_alpha
        self.continuous_columns = continuous_columns
        self.categorical_columns = categorical_columns

    def fit(self, X, a):
        """Fit the propensity filter

        This fits a `Support` object for every column depending on its dtype.
        It also calculates the scales of the original data.

        Args:
            X (pd.DataFrame): covariates DataFrame
            a (pd.Series): treatment assignment Series

        Returns:
            UnivariateBoundingBox: Fitted positivity filter
        """
        self.treatment_support_ = {
            c: self.fit_column(X[a == 1][c]) for c in X.columns}
        self.control_support_ = {c: self.fit_column(
            X[a == 0][c]) for c in X.columns}
        self.joint_support_ = {c: self.treatment_support_[c].intersection(
            self.control_support_[c]) for c in X.columns}
        self.scales_ = self._calc_scales(X, a)
        return self

    def fit_column(self, Xcol):
        """Fit an individual column

        Args:
            Xcol (pd.Series|pd.DataFrame): a single column of data

        Returns:
            Support: a fitted Support object
        """
        if self._is_column_Continuous(Xcol):
            if self.quantile_alpha is None:
                return ContinuousSupport().fit(Xcol)
            else:
                return QuantileContinuousSupport(alpha=self.quantile_alpha).fit(Xcol)
        else:
            return CategoricalSupport().fit(Xcol)


    def predict(self, X, a=None):
        """Predict whether the sample is in the support for all variables

        Note that the treatment assignment vector `a` is not used with this
        method. Every sample must be in every joint support to be considered
        in the overlapped set, regardless of its treatment value.

        Args:
            X (pd.DataFrame): covariates
            a (pd.Series): treatment assignment 

        Returns:
            pd.Series: a binary series of length `X.shape[0]` with True for each
                sample determined to be in the support else False

        Raises:
            NotFittedError: if not fitted
        """
        self.assert_is_fitted()
        in_overlap_for_column = {c: s.predict for c,
                                 s in self.joint_support_.items()}
        return X.transform(in_overlap_for_column).apply(all, axis=1)

    @property
    def supports_table_(self):
        """DataFrame summarizing the fitted support variables.

        Raises:
            NotFittedError: if not fitted
        """
        self.assert_is_fitted()
        d=dict()
        for i,j,k,l in zip(
                self.treatment_support_,
                self.treatment_support_.values(),
                self.control_support_.values(),
                self.joint_support_.values()):
            d[i] = dict(treatment=j, control=k, joint=l)
        return pd.DataFrame(d).T


    @property
    def scaled_supports_table_(self):
        """DataFrame summarizing the fitted support variables in rescaled units.

        Raises:
            NotFittedError: if not fitted
        """
        self.assert_is_fitted()
        d=dict()
        def rescaler(column_name):
            def f(support):
                if isinstance(support.support,list):
                    scaled_support =  [support.support[0]/self.scales_[column_name], support.support[1]/self.scales_[column_name]]
                    return type(support)(support = scaled_support)
                else:
                    return support
            return f
        for i,j,k,l in zip(
                self.treatment_support_,
                self.treatment_support_.values(),
                self.control_support_.values(),
                self.joint_support_.values()):
            rs = rescaler(i)
            d[i] = dict(treatment=rs(j), control=rs(k), joint=rs(l))
        return pd.DataFrame(d).T

    def assert_is_fitted(self):
        """Check if filter is fitted

        Raises:
            NotFittedError: if not fitted
        """
        if not hasattr(self, "joint_support_"):
            raise NotFittedError("You must run `fit` first")

    @staticmethod
    def _calc_scales( X, a):
        n0 = X[a == 0].shape[0] - 1
        n1 = X[a == 0].shape[0] - 1
        return np.sqrt((n0*X[a == 0].var() + n1*X[a == 1].var())/(n0 + n1))

    def _is_column_Continuous(self, Xcol):
        if Xcol.name in self.categorical_columns:
            return False
        if Xcol.name in self.continuous_columns:
            return True
        if Xcol.dtype == float:
            return True
        return False



import unittest

import pandas as pd
from sklearn.linear_model import LogisticRegression
from sklearn.exceptions import NotFittedError

from causallib.positivity import (
    Trimming, UnivariateBoundingBox, Matching
)
from causallib.positivity.univariate_bbox import (
    QuantileContinuousSupport, Support, ContinuousSupport, CategoricalSupport
)
from causallib.positivity.datasets.positivity_data_simulator import (
    make_1d_overlap_data, make_1d_normal_distribution_overlap_data, make_random_y_like
)


class TestTrimming(unittest.TestCase):

    # TODO: check the minimization criterion - following Crump's paper

    @classmethod
    def setUpClass(cls):
        # Propensity model without regularization
        cls.trimming = Trimming(
            learner=LogisticRegression(solver="sag", penalty=None),
            threshold="auto"
        )

    def test_is_fitted(self):
        X, a = make_1d_normal_distribution_overlap_data()
        self.trimming.fit(X, a)
        self.assertTrue(hasattr(self.trimming.learner, "coef_"))
        self.assertTrue(isinstance(self.trimming.threshold_, float))

    def test_not_trimming_1d_same_distribution(self):
        # Treated and control are drawn from the same distribution with
        # probability to be treated = 0.5
        X, a = make_1d_normal_distribution_overlap_data()
        [X_trimmed, a_trimmed] = self.trimming.fit_transform(X, a)
        self.assertEqual(a_trimmed.shape[0], a.shape[0])

        # Treated and control are drawn from the same distribution with
        # probability to be treated = 0.2
        X, a = make_1d_normal_distribution_overlap_data(probability_treated=0.2)
        [X_trimmed, a_trimmed] = self.trimming.fit_transform(X, a)
        self.assertEqual(a_trimmed.shape[0], a.shape[0])

    def test_trimming_different_distribution(self):
        X, a = make_1d_normal_distribution_overlap_data(
            treatment_params=(-2, 2),
            control_params=(2, 1),
            probability_treated=0.4,
            n_samples=1000
        )
        [X_trimmed_crump, a_trimmed_crump] = self.trimming.fit_transform(X, a)
        n_untrimmed_crump = a_trimmed_crump.shape[0]
        with self.subTest("Trimmed samples with crump"):
            self.assertTrue(n_untrimmed_crump < a.shape[0])

        n_untrimmed_big_threshold = \
            self.trimming.predict(X, a, threshold=0.25).sum()
        with self.subTest("Trimmed samples with large threshold"):
            self.assertTrue(n_untrimmed_big_threshold < a.shape[0])

        with self.subTest("Test crump has more sample than large threshold"):
            self.assertTrue(n_untrimmed_big_threshold < n_untrimmed_crump)

    def test_pipeline(self):
        X, a = make_1d_normal_distribution_overlap_data()

        with self.subTest("Test fit with propensity model"):
            self.trimming.fit(X, a)
            self.assertTrue(True)  # Dummy assert for not thrown exception

        with self.subTest("Test predict of the filtered data"):
            self.trimming.predict(X, a)
            self.assertTrue(True)  # Dummy assert for not thrown exception

        with self.subTest("Test transform"):
            self.trimming.transform(X, a)
            self.assertTrue(True)  # Dummy assert for not thrown exception

    def test_hardcoded_threshold_initialization_chooses_no_method(self):
        self.trimming = Trimming(threshold=0.1)
        self.assertFalse(hasattr(self.trimming, 'threshold'))

    def test_default_initialization_chooses_crump(self):
        self.trimming = Trimming()
        self.assertEqual(self.trimming.threshold, "crump")



def inbounds(bounds):
    return lambda x: x >= bounds[0] and x <= bounds[1]


class TestUnivariateBBox(unittest.TestCase):
    def assertHasAttr(self, obj, intendedAttr):
        testBool = hasattr(obj, intendedAttr)
        self.assertTrue(testBool)

    @classmethod
    def setUpClass(cls):
        cls.bbox = UnivariateBoundingBox()
        cls.treatment_bounds_1d = (0, 75)
        cls.control_bounds_1d = (0, 75)
        cls.overlap_data_discrete_integer = make_1d_overlap_data(
            treatment_bounds=cls.treatment_bounds_1d,
            control_bounds=cls.control_bounds_1d)

        cls.treatment_params = (0, 1)
        cls.control_params = (.5, 1)
        cls.overlap_data_normal = make_1d_normal_distribution_overlap_data(
            treatment_params=cls.treatment_params, control_params=cls.control_params)

    def test_predict_is_series(self):
        X, a = self.overlap_data_discrete_integer
        self.bbox.fit(X, a)
        pred = self.bbox.predict(X, a)
        self.assertIsInstance(pred, pd.Series)

    def test_fit_makes_support_(self):
        X, a = self.overlap_data_discrete_integer
        self.bbox.fit(X, a)
        self.assertHasAttr(self.bbox, "joint_support_")

    def test_fit_predict_trues_are_correct_integers(self):
        X, a = self.overlap_data_discrete_integer
        pred = self.bbox.fit_predict(X, a)
        self.assertGreaterEqual(
            X[pred].squeeze().min(),
            max(self.treatment_bounds_1d[0], self.control_bounds_1d[0]))
        self.assertLessEqual(
            X[pred].squeeze().max(),
            min(self.treatment_bounds_1d[1], self.control_bounds_1d[1]))

    def test_fit_predict_falses_are_correct_integers(self):
        X, a = self.overlap_data_discrete_integer
        pred = self.bbox.fit_predict(X, a)
        lr1 = self.control_bounds_1d
        lr2 = self.treatment_bounds_1d

        Xinbounds = X[~pred].squeeze().apply(
            lambda x: inbounds(lr1)(x) and inbounds(lr2)(x))
        self.assertFalse(any(Xinbounds))

    def test_fit_predict_trues_are_correct_floats(self):
        X, a = self.overlap_data_normal
        pred, in_both_bounds = self._check_in_bounds(X, a)
        self.assertTrue(all(X[pred].squeeze().apply(in_both_bounds)))

    def test_fit_predict_falses_are_correct_floats(self):
        X, a = self.overlap_data_normal
        pred, in_both_bounds = self._check_in_bounds(X, a)
        self.assertFalse(any(X[~pred].squeeze().apply(in_both_bounds)))

    def _check_in_bounds(self, X, a):
        pred = self.bbox.fit_predict(X, a)
        X = X.squeeze()
        lr1 = self.bbox.fit_column(X[a == 0]).support
        lr2 = self.bbox.fit_column(X[a == 1]).support
        def in_both_bounds(x): return inbounds(lr1)(x) and inbounds(lr2)(x)
        return pred, in_both_bounds

    def test_continuous_support(self):
        X, a = self.overlap_data_normal
        self.bbox.quantile_alpha = None
        X = X.squeeze()

        self._column_matches_continuous_support(X[a == 0])
        self._column_matches_continuous_support(X[a == 1])

    def _column_matches_continuous_support(self, Xt_col):
        Xtsupp = self.bbox.fit_column(Xt_col)
        self.assertListEqual([min(Xt_col), max(Xt_col)], Xtsupp.support)

    def test_continuous_fit(self):
        X, a = self.overlap_data_normal
        self.bbox.quantile_alpha = None
        self.bbox.fit(X, a)
        self._check_manual_minmax_to_bbox_support(X, a, 0)
        self._check_manual_minmax_to_bbox_support(X, a, 1)

    def _check_manual_minmax_to_bbox_support(self, X, a, treatment_value):
        c = X.columns[0]
        minmax = [X[a == treatment_value]
                  [c].min(), X[a == treatment_value][c].max()]
        bbox_support = self.bbox.control_support_[
            c] if treatment_value == 0 else self.bbox.treatment_support_[c]
        self.assertListEqual(minmax, bbox_support.support)

    def test_continuous_support_quantile(self):
        self.bbox.quantile_alpha = 0.1

        X, a = self.overlap_data_normal
        X = X.squeeze()

        q = [self.bbox.quantile_alpha/2, 1 - (self.bbox.quantile_alpha/2)]

        self._fit_column_and_support_matches_quantile(X[a == 1], q)
        self._fit_column_and_support_matches_quantile(X[a == 0], q)

    def _fit_column_and_support_matches_quantile(self, Xt_col, q):
        Xtsupp = self.bbox.fit_column(Xt_col)
        self.assertListEqual(list(Xt_col.quantile(q)), Xtsupp.support)

    def test_continuous_fit_quantile(self):
        X, a = self.overlap_data_normal
        self.bbox.quantile_alpha = 0.1
        self.bbox.fit(X, a)
        self._check_manual_quantile_to_bbox_support(X, a, 0)
        self._check_manual_quantile_to_bbox_support(X, a, 1)

    def _check_manual_quantile_to_bbox_support(self, X, a, treatment_value):
        c = X.columns[0]
        X = X.squeeze()
        minmax = [
            X[a == treatment_value].quantile(self.bbox.quantile_alpha/2),
            X[a == treatment_value].quantile(1 - (self.bbox.quantile_alpha/2))]
        bbox_support = self.bbox.control_support_[
            c] if treatment_value == 0 else self.bbox.treatment_support_[c]
        self.assertListEqual(minmax, bbox_support.support)

    def test_intersecting_nonmatching_supports_raises_value_error(self):
        Xn, an = self.overlap_data_normal
        Xi, ai = self.overlap_data_discrete_integer
        cts_support = self.bbox.fit_column(Xn.squeeze())
        int_support = self.bbox.fit_column(Xi.squeeze())
        with self.assertRaises(ValueError):
            cts_support.intersection(int_support)
        with self.assertRaises(ValueError):
            int_support.intersection(cts_support)

    def test_intersecting_continuous_and_quantile_supports(self):
        Xn, an = self.overlap_data_normal
        Xn = Xn.squeeze()
        cts_support = ContinuousSupport().fit(Xn)
        quantile_support = QuantileContinuousSupport(alpha=0.1).fit(Xn)
        joint_support = cts_support.intersection(quantile_support)
        joint_support2 = quantile_support.intersection(cts_support)
        self.assertEqual(joint_support.support, quantile_support.support)
        self.assertEqual(joint_support2.support, quantile_support.support)

    def test_supports_table(self):
        Xn, an = self.overlap_data_normal
        self.bbox.fit(Xn, an)
        self.assertListEqual(
            sorted(list(self.bbox.supports_table_.columns)),
            sorted(["control", "joint", "treatment", ]))
        self.assertListEqual(sorted(list(self.bbox.supports_table_.index)),
                             sorted(Xn.columns))
        if pd.__version__ > "2.1":
            # TODO: keep some support for pandas <2.1.0 and python 3.8. remove in the future.
            self.bbox.supports_table_.map(lambda x: isinstance(x, Support))
        else:
            self.bbox.supports_table_.applymap(lambda x: isinstance(x, Support))

    def test_subtract_continuous_support_when_equal(self):
        csupport1 = ContinuousSupport(support=[1, 2])
        csupport2 = ContinuousSupport(support=[1, 2])
        self.assertListEqual(csupport1 - csupport2, [0, 0])

    def test_substract_continuous_support_when_unequal(self):
        csupport1 = ContinuousSupport(support=[1, 2])
        csupport2 = ContinuousSupport(support=[2, 5])
        self.assertListEqual(csupport1 - csupport2, [-1, -3])

    def test_subtract_discrete_support_when_equal(self):
        dsupport1 = CategoricalSupport(support=[1, 2, 3])
        dsupport2 = CategoricalSupport(support=[2, 3, 1])
        self.assertFalse(dsupport1 - dsupport2)

    def test_subtract_discrete_support_when_unequal(self):
        dsupport1 = CategoricalSupport(support=[2])
        dsupport2 = CategoricalSupport(support=[2, 3, 1])
        self.assertSetEqual(dsupport1 - dsupport2, {3, 1})

    def test_is_fitted_check(self):
        self.bbox = UnivariateBoundingBox()
        with self.assertRaises(NotFittedError):
            self.bbox.assert_is_fitted()

    def test_str_representation_for_continuous_support(self):
        self.assertEqual(str(ContinuousSupport([0, 1])), "support: [0, 1]")
        self.assertEqual(str(ContinuousSupport()), "support: no support")

    def test_str_representation_for_discrete_support(self):
        self.assertEqual(str(CategoricalSupport([0, 1])), "support: {0, 1}")
        self.assertEqual(str(CategoricalSupport()), "support: no support")

    def test_can_select_continuous_columns(self):
        X, a = self.overlap_data_discrete_integer
        self.bbox = UnivariateBoundingBox(continuous_columns=["X1"])
        self.bbox.fit(X, a)
        self.assertIsInstance(self.bbox.joint_support_[
                              "X1"], ContinuousSupport)

    def test_can_select_categorical_columns(self):
        X, a = self.overlap_data_discrete_integer
        self.bbox = UnivariateBoundingBox(categorical_columns=["X1"])
        self.bbox.fit(X, a)
        self.assertIsInstance(self.bbox.joint_support_[
                              "X1"], CategoricalSupport)

    def test_non_overlapping_intervals_are_none(self):
        s1 = ContinuousSupport().fit([1, 2, 3])
        s2 = ContinuousSupport().fit([4, 5, 6])
        s3 = s1.intersection(s2)
        self.assertIsNone(s3.support)
        s2 = ContinuousSupport().fit([1, 2, 3])
        s1 = ContinuousSupport().fit([4, 5, 6])
        s3 = s1.intersection(s2)
        self.assertIsNone(s3.support)

    def test_overlapping_with_none_is_none(self):
        s1 = ContinuousSupport(support=None)
        s2 = ContinuousSupport().fit([1,2,3])
        s3 = s1.intersection(s2)
        self.assertIsNone(s3.support)

    def test_scaled_support_table(self):
        Xn, an = self.overlap_data_normal
        self.bbox.fit(Xn, an)
        supports_table = self.bbox.supports_table_["treatment"]
        scaled_supports_table = self.bbox.scaled_supports_table_["treatment"]
        something_asserted = False
        for col in Xn.columns:
            if isinstance(supports_table[col].support,list):
                s0,s1 = supports_table[col].support
                ss0,ss1 = scaled_supports_table[col].support
                s = self.bbox.scales_[col]
                self.assertAlmostEqual(s0/s,ss0)
                self.assertAlmostEqual(s1/s,ss1)
                something_asserted = True
        self.assertTrue(something_asserted, "No asserts made, check data")


class TestMatching(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.matching = Matching()

    def test_1d_euclidean_caliper_1_match(self):
        X, a= make_1d_overlap_data(treatment_bounds=(0, 75),
                                      control_bounds=(25, 100))
        self.matching.matching_transformer.set_params(metric="euclidean", caliper=1.0)
        predict = self.matching.fit_predict(X, a)
        Xp, ap= self.matching.transform(X, a,)
        self.assertEqual(sum(predict), 100)
        self.assertEqual(sum(predict), Xp.shape[0])
        self.assertEqual(Xp.values.min(), 25)
        self.assertEqual(Xp.values.max(), 74)

    def test_transform_multiple_outputs(self):
        X, a= make_1d_overlap_data(treatment_bounds=(0, 75),
                                      control_bounds=(25, 100))
        y = make_random_y_like(a)
        t = make_random_y_like(a)

        self.matching.matching_transformer.set_params(metric="euclidean", caliper=1.01)
        self.matching.fit(X, a)

        Xm, am = self.matching.transform(X, a)
        unique_sample_counts = {i.shape[0] for i in (Xm, am)}
        self.assertEqual(len(unique_sample_counts), 1)

        Xm, am, ym= self.matching.transform(X, a, y)
        unique_sample_counts = {i.shape[0] for i in (Xm, am, ym)}
        self.assertEqual(len(unique_sample_counts), 1)

        Xm, am, ym, tm = self.matching.transform(X, a, y, t)
        unique_sample_counts = {i.shape[0] for i in (Xm, am, ym, tm)}
        self.assertEqual(len(unique_sample_counts), 1)



import pandas as pd
import numpy as np


def make_1d_overlap_data(treatment_bounds=(0, 75),
                         control_bounds=(25, 100)):
    """Generate 1d overlap data with integer covariates

    Args:
        treatment_bounds (tuple, optional): Bounds for covariates in treatment
            group. Defaults to (0, 75).
        control_bounds (tuple, optional): Bounds for covariates in control
            group. Defaults to (25, 100).

    Returns:
        X (pd.DataFrame), a (pd.Series): covariate and treatment assignment
    """

    X_treatment = np.arange(*treatment_bounds)
    a_treatment = np.ones_like(X_treatment)

    X_control = np.arange(*control_bounds)
    a_control = np.zeros_like(X_control)

    X = pd.DataFrame(data=np.hstack((X_treatment, X_control)), columns=["X1"])
    a = pd.Series(data=np.hstack((a_treatment, a_control)), name="treatment")

    return X, a


def make_1d_normal_distribution_overlap_data(treatment_params=(0, 1),
                                             control_params=(0, 1),
                                             probability_treated=0.5,
                                             n_samples=400,
                                             random_seed=1234):
    """
    Args:
        treatment_params (tuple): loc and scale parameter of normal distribution
        control_params (tuple): loc and scale parameter of normal distribution
        probability_treated (float):
        n_samples (int):
        random_seed (int):
    Returns:
        X (pd.DataFrame), a (pd.Series):
    """
    n_treated = int(round(n_samples * probability_treated))
    n_control = n_samples - n_treated

    a_control = np.zeros(n_control)
    np.random.seed(random_seed)
    X_control = np.random.normal(
        loc=control_params[0], scale=control_params[1], size=n_control)

    a_treatment = np.ones(n_treated)
    np.random.seed(random_seed + 1)
    X_treatment = np.random.normal(
        loc=treatment_params[0], scale=treatment_params[1], size=n_treated)

    X = pd.DataFrame(data=np.hstack((X_treatment, X_control)), columns=["X"])
    a = pd.Series(data=np.hstack((a_treatment, a_control)), name="treatment")
    return X, a


def make_multivariate_normal_data(
        treatment_params=([0, 0], [[2, 1], [1, 2]]),
        control_params=([2, 2], [[1, 0], [0, 1]]),
        probability_treated=0.5,
        n_samples=400,
        random_seed=1234):
    """
    Args:
        treatment_params (tuple): loc and scale parameter of multivariate normal distribution
        control_params (tuple): loc and scale parameter of multivariate normal distribution
        probability_treated (float): the probability to be in the treatment group
        n_samples (int): number of samples for the full data set
        random_seed (int): each group receive different seeding
    Returns:
        X (pd.DataFrame), a (pd.Series):
    """
    n_treated = int(round(n_samples * probability_treated))
    n_control = n_samples - n_treated

    a_control = np.zeros(n_control)
    np.random.seed(random_seed)
    X_control = np.random.multivariate_normal(control_params[0],
                                              control_params[1],
                                              n_control)
    a_treatment = np.ones(n_treated)
    np.random.seed(random_seed + 1)
    X_treatment = np.random.multivariate_normal(treatment_params[0],
                                                treatment_params[1],
                                                n_treated)

    X = pd.DataFrame(data=np.vstack((X_treatment, X_control)))
    a = pd.Series(data=np.hstack((a_treatment, a_control)), name="treatment")
    return X, a


def make_random_y_like(a, random_seed=1234):
    np.random.seed(random_seed)
    y = pd.Series(data=np.random.random_sample(a.shape), name="outcome")
    return y



import unittest

import pandas as pd

from sklearn.linear_model import LogisticRegression
from sklearn.datasets import make_classification
from sklearn.utils import Bunch
from sklearn.metrics import get_scorer

from causallib.estimation import IPW
from causallib.model_selection import GridSearchCV

from causallib.contrib.sklearn_scorer_wrapper import SKLearnScorerWrapper


class TestSKLearnScorerWrapper(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        N = 500
        X, a = make_classification(
            n_samples=N,
            n_features=5,
            n_informative=5,
            n_redundant=0,
            random_state=42,
        )
        X = pd.DataFrame(X)
        a = pd.Series(a)
        cls.data = Bunch(X=X, a=a, y=a)

        learner = LogisticRegression()
        ipw = IPW(learner)
        ipw.fit(X, a)
        # cls.learner = learner
        cls.estimator = ipw

    def test_agreement_with_sklearn(self):
        scorer_names = [
            "accuracy",
            "average_precision",
            "neg_brier_score",
            "f1",
            "neg_log_loss",
            "precision",
            "recall",
            "roc_auc",
        ]
        for scorer_name in scorer_names:
            with self.subTest(f"Test scorer {scorer_name}"):
                scorer = get_scorer(scorer_name)
                score = scorer(self.estimator.learner, self.data.X, self.data.a)

                causallib_adapted_scorer = SKLearnScorerWrapper(scorer)
                causallib_score = causallib_adapted_scorer(
                    self.estimator, self.data.X, self.data.a, self.data.y
                )

                self.assertAlmostEqual(causallib_score, score)

    def test_hyperparameter_search_model(self):
        scorer = SKLearnScorerWrapper(get_scorer("roc_auc"))
        param_grid = dict(
            clip_min=[0.2, 0.3],
            learner__C=[0.1, 1],
        )
        model = GridSearchCV(
            self.estimator,
            param_grid=param_grid,
            scoring=scorer,
            cv=3,
        )
        model.fit(self.data.X, self.data.a, self.data.y)

        score = scorer(model, self.data.X, self.data.a, self.data.y)
        self.assertGreaterEqual(score, model.best_score_)



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on Oct 26, 2021

"""
from collections import deque
from copy import deepcopy
from typing import Union

import numpy as np
import pandas as pd
from statsmodels.stats.multitest import multipletests
from scipy.stats import fisher_exact, chi2_contingency
from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.dummy import DummyClassifier

from causallib.contrib.bicause_tree.overlap_utils import prevalence_symmetric, OverlapViolationEstimator
from causallib.utils.general_tools import get_iterable_treatment_values
from causallib.estimation.base_estimator import PopulationOutcomeEstimator, IndividualOutcomeEstimator
from causallib.estimation import MarginalOutcomeEstimator
from causallib.metrics.weight_metrics import calculate_covariate_balance


def default_stopping_criterion(tree, X: pd.DataFrame, a: pd.Series):
    """
    Args:
        tree (BalancingTree): the current node (tree) being considered
        X (pd.DataFrame): The feature matrix
        a (pd.Series): the treatment assignment vector
    Returns:
        (boolean): is the stopping criterion met
    """
    # returns True if we must stop

    asmds = calculate_asmds(X, a)
    no_exploitable_asmds = asmds.isna().all()
    # asmds can be NaN if the standard deviation is null
    if no_exploitable_asmds:
        return True
    _is_asmd_threshold_reached = asmds.max() <= tree.asmd_violation_threshold
    _is_min_treat_group_size_reached = min(a.value_counts()) <= tree.min_treat_group_size
    _is_min_split_size_reached = X.shape[0] <= tree.min_split_size
    _is_max_depth_tree_reached = tree.max_depth <= 1

    criteria = any([
        _is_asmd_threshold_reached,
        _is_min_treat_group_size_reached,
        _is_min_split_size_reached,
        _is_max_depth_tree_reached
    ])

    return criteria


def calculate_asmds(X: pd.DataFrame, a: pd.Series):
    epsilon = np.finfo(float).resolution
    treatment_values = get_iterable_treatment_values(None, a)
    if len(treatment_values) < 2:
        return pd.Series(np.nan, index=X.columns)
    X0 = X.loc[a == treatment_values[0]]
    X1 = X.loc[a != treatment_values[0]]
    asmds = (X0.mean() - X1.mean()) / ((np.sqrt(X0.var() + X1.var())) + epsilon)
    asmds = asmds.abs()
    return asmds


class PropensityBICauseTree:

    def __init__(
        self,
        learner=DummyClassifier(strategy='prior'),
        **kwargs
    ) -> None:
        """
        A propensity model based on a BICause Tree,
        scikit-learn compatible

        Args:
            learner (BaseEstimator, ClassifierMixin): The propensity model
            **kwargs: arguments for `BalancingTree`. See its docstring for details.
        """
        self.learner = learner
        self.tree = BalancingTree(**kwargs)

    def fit(self, X, a):
        """Fits a PropensityBICauseTree propensity model.

        Compatible with scikit-learn classifiers,
        which can be combined with causallib's models
        - e.g., IPW(PropensityBICauseTree())

        Args:
            X (pd.DataFrame): The feature matrix
            a (pd.Series): The treatment assignment vector

        Returns:
            PropensityBICauseTree
        """
        self.tree.fit(X, a)
        self.fit_treatment_models(X, a)
        return self

    def fit_treatment_models(self, X, a):
        """Fits the probability estimator in the leaf nodes.

        Args:
            X (pd.DataFrame): The feature matrix
            a (pd.Series): The treatment assignment vector

        Returns:
            dict[int, BaseEstimator]: a mapping between leaf index and
                the corresponding probability estimator
        """
        self.treatment_values_ = np.unique(a)

        assignment = self.tree.apply(X)
        all_leaves = assignment.unique()
        node_models = {}
        for node in all_leaves:
            cur_node_mask = assignment == node
            current_node_X = X.loc[cur_node_mask]
            current_node_a = a.loc[cur_node_mask]

            model = self.learner.fit(current_node_X, current_node_a)
            node_models[node] = deepcopy(model)

        self.node_models_ = node_models
        return node_models

    def predict_proba(self, X):
        """ Predict the individual treatment probabilities.

        Args:
            X (pd.DataFrame): The feature matrix

        Returns:
            np.array: A matrix of individual treatment probabilities,
                (n_samples, n_treatment_values)
        """
        node_assignment = self.tree.apply(X)
        res = []
        for node in np.unique(node_assignment.values):
            curr_node_index = node_assignment == node
            curr_X = X.loc[curr_node_index]
            curr_model = self.node_models_[node]

            probability = curr_model.predict_proba(curr_X)
            probability = pd.DataFrame(
                probability,
                columns=curr_model.classes_,
                index=curr_X.index,
            )
            res.append(probability)

        res = pd.concat(res)
        res = res.fillna(0.0)  # If classes don't fully overlap across leaves
        res = res.loc[X.index]
        res = res.to_numpy()

        return res


class BICauseTree(IndividualOutcomeEstimator):

    def __init__(
        self,
        outcome_model: Union[IndividualOutcomeEstimator, PopulationOutcomeEstimator] = None,
        individual=False,
        asmd_violation_threshold=0.1,
        min_leaf_size=0,
        min_treat_group_size=0,
        min_split_size=0,
        max_depth=10,
        stopping_criterion=default_stopping_criterion,
        max_splitting_values=50,
        multiple_hypothesis_test_method='holm',
        multiple_hypothesis_test_alpha=0.1,
        positivity_filtering_method=prevalence_symmetric,
        positivity_filtering_kwargs=None,
    ) -> None:
        """A causal model effect estimator built on top of a tree recursively stratifying the covariate space to
        balance between treated and untreated.

        Args:
            outcome_model (Union[IndividualOutcomeEstimator, PopulationOutcomeEstimator]):
                An outcome model for generating counterfactual predictions at each leaf node of the tree.
                Defaults to a simple `MarginalOutcomeEstimator` that just takes the average outcome for each treatment
                group in each leaf. However, it may also be any arbitrary outcome model to further adjust for the
                covariates (that the tree might leave some residual bias in the stratification).
            individual (bool): If True (and if `outcome_model` has `estimate_individual_outcomes`)
                will generate individual-level predictions for observations within each leaf.
                Otherwise, each observation takes the value of the average outcome in that leaf
                (using the `estimate_population_outcomes` method).
            asmd_violation_threshold (float): The value of Absolute Standardized Mean Difference below which a subgroup
                is considered balanced.
            min_leaf_size (int): The minimum number of samples required to split an internal node
            min_treat_group_size (int): The minimum number of samples in all treatment groups
                required to split an internal node.
            min_split_size (int): The minimum number of samples required to split an internal node.
            max_depth (int): The maximum depth of the tree.
                Will be updated for each level of nodes as the tree grows.
            stopping_criterion (callable): A function that takes the node/subtree as well as the data (`X`, `a`)
                and returns a boolean `True` if to stop splitting the tree and `False` if to continue splitting.
            max_splitting_values (int): The maximal number of unique values to consider when splitting a single feature
            multiple_hypothesis_test_method: The method for correcting p-values in multiple hypotheses testing.
                Should be compatible with statsmodels' `multipletests`.
            multiple_hypothesis_test_alpha (float): The alpha value for correcting p-values in
                multiple hypotheses testing.
                Should be compatible with statsmodels' `multipletests`.
            positivity_filtering_method (callable):
                A function that takes the current node (or subtree) as well as
                the arbitrary kwargs from `positivity_filtering_kwargs` and
                returns a list of the leaves/nodes' indices that do *not* violate positivity.
            positivity_filtering_kwargs: Keyword arguments to call the `positivity_filtering_method` with.
        """
        super().__init__(learner=None)

        if outcome_model is None:
            self.outcome_model = MarginalOutcomeEstimator(None)
        else:
            self.outcome_model = deepcopy(outcome_model)

        if positivity_filtering_kwargs is None:
            positivity_filtering_kwargs = {'alpha': 0.1}

        self.individual = individual
        self.tree = BalancingTree(
            asmd_violation_threshold=asmd_violation_threshold,
            min_leaf_size=min_leaf_size,
            min_treat_group_size=min_treat_group_size,
            min_split_size=min_split_size, max_depth=max_depth,
            stopping_criterion=stopping_criterion,
            max_splitting_values=max_splitting_values,
            multiple_hypothesis_test_method=multiple_hypothesis_test_method,
            multiple_hypothesis_test_alpha=multiple_hypothesis_test_alpha,
            positivity_filtering_method=positivity_filtering_method,
            positivity_filtering_kwargs=positivity_filtering_kwargs
        )

    def fit(self, X, a, y, sample_weight=None):
        """ Build the BICauseTree partition

        Args:
            X (pd.DataFrame): The feature matrix of all samples
            a (pd.Series): The treatment assignments
            y (pd.Series): The outcome values
            sample_weight: IGNORED

        Returns:
            (class) BICauseTree
        """
        self.tree.fit(X, a)
        self.fit_outcome_models(X, a, y)
        return self

    def apply(self, X):
        """ Get node assignments based on BICauseTree partition

        Args:
            X (pd.DataFrame): The feature matrix of all samples

        Returns:
            (pd.Series) A vector of node indices indexed according to X
        """
        assignment = self.tree.apply(X)
        return assignment

    def estimate_population_outcome(
        self,
        X, a, y=None,
        discard_violating_samples=True,
        agg_func="mean",
    ):
        """ Estimates outcomes at a population-level and assigns them to individuals in X

        Args:
            X (pd.DataFrame): The feature matrix of all samples
            a (pd.Series): The treatment assignments
            y (pd.Series): The outcome values
            discard_violating_samples (boolean): whether to drop the NA in the individual outcomes
            agg_func : aggregation function to go from individual to population outcome estimates

        Returns:
            A vector of potential outcomes indexed according to X
        """
        individual_cf = self.estimate_individual_outcome(X, a, y)
        if discard_violating_samples:
            individual_cf = individual_cf.dropna()
        if individual_cf.isnull().any().any():
            treatment_values = len(individual_cf.columns)
            pop_outcome = pd.Series(np.nan, index=range(0, treatment_values))
        else:
            pop_outcome = individual_cf.apply(self._aggregate_population_outcome, args=(agg_func,))
        return pop_outcome

    def estimate_individual_outcome(self, X, a, y=None, same_dim_as_input=True):
        """estimate individual-level counterfactual predictions.

        if `self.individual is True` and `self.outcome_model` has `estimate_individual_outcome()`
        then each observation will get a unique counterfactual value.
        otherwise, each individual gets the average prediction of its node,
        and this function is a way to transform it to a shape of predictions-per-observation.

        Args:
            X (pd.DataFrame): The feature matrix of all samples
            a (pd.Series): The treatment assignments
            y (pd.Series): The outcome values
            same_dim_as_input (boolean): whether to return nan values for
                positivity-violating observations or exclude them

        Returns:
            pd.DataFrame: A matrix of individual outcomes indexed according to X
        """
        node_assignment = self.apply(X)
        all_df = []
        for node in np.unique(node_assignment.values):

            curr_node_index = node_assignment == node
            curr_X = X.loc[curr_node_index]
            curr_a = a[curr_node_index]
            curr_y = y[curr_node_index] if y is not None else None
            curr_model = self.node_models_[node]

            if self.individual:
                if hasattr(curr_model, "estimate_individual_outcome"):
                    outcomes = curr_model.estimate_individual_outcome(curr_X, curr_a)
                else:
                    raise AttributeError(
                        "The `outcome_model` provided can't estimate individual outcomes, "
                        "even though `self.individual` is set to True."
                        "Please change the `outcome_model` or set `self.individual` to False."
                    )
            else:
                outcomes = curr_model.estimate_population_outcome(curr_X, curr_a, curr_y)
                # we duplicate the outcome values as the len of the vector to concat to the df
                outcomes = pd.DataFrame(outcomes).transpose()
                outcomes = pd.concat([outcomes] * len(curr_a), ignore_index=True)
                outcomes = outcomes.set_index(curr_X.index)

            all_df.append(outcomes)

        # concat all df together
        df = pd.concat(all_df)
        df = df.loc[X.index]

        if not same_dim_as_input:
            df = df.dropna()

        return df

    def fit_outcome_models(self, X, a, y):
        """ Fits causal models to the nodes of a fitted (already grown) tree.

        Args:
            X (pd.DataFrame): The feature matrix of all samples
            a (pd.Series): The treatment assignments
            y (pd.Series): The outcome values

        Returns:
            dict[int, Union[IndividualOutcomeEstimator, PopulationOutcomeEstimator]
        """
        positivity_violating_leaves = self.tree.get_positivity_violation_status()
        assignment = self.tree.apply(X)
        total_nodes = assignment.unique()

        node_models = {}
        for node in total_nodes:

            has_positivity_violation = positivity_violating_leaves[node]
            idx = assignment[assignment == node].index
            current_node_X = X.loc[idx]
            current_node_a = a.loc[idx]
            current_node_y = y.loc[idx]

            model = OverlapViolationEstimator() if has_positivity_violation else deepcopy(self.outcome_model)
            model.fit(current_node_X, current_node_a, current_node_y)
            node_models[node] = deepcopy(model)

        self.node_models_ = node_models
        return node_models

    def explain(self, X, a, split_condition: str = None):
        """Create a list of data frames summarizing the decision tree and the marginal effect.
            Each data-frame represents a leaf in the tree, and the list represents the tree itself.
            Each data frame exhibits several summary statistics about the path from the root to the
            leaf, including the maximal asmd at that level and the marginal outcome value.

        Args:
            X (pd.DataFrame): The feature data. Assumed to be of the same column
            structure as the training data
            a (pd.Series): The treatment assignment vector
            split_condition (str): The string representing the first condition. Default to 'All'
            to which the split explanations are added

        Returns:
            List[pd.DataFrame]: The list representing the tree, holding a data-frame for every leaf.
        """
        return self.tree.explain(X, a, split_condition)


class BalancingTree:
    def __init__(
        self,
        asmd_violation_threshold=0.1,
        min_leaf_size=0,
        min_treat_group_size=0,
        min_split_size=0,
        max_depth=10,
        stopping_criterion=default_stopping_criterion,
        max_splitting_values=50,
        multiple_hypothesis_test_method='holm',
        multiple_hypothesis_test_alpha=0.1,
        positivity_filtering_method=None,
        positivity_filtering_kwargs=None,
        _parent_=None
    ):
        """A tree node that recursively stratifies the covariate space based on ASMD value and other constraints.

        Args:
            asmd_violation_threshold (float): The value of Absolute Standardized Mean Difference below which a subgroup
                is considered balanced.
            min_leaf_size (int): The minimum number of samples required to split an internal node
            min_treat_group_size (int): The minimum number of samples in all treatment groups
                required to split an internal node.
            min_split_size (int): The minimum number of samples required to split an internal node.
            max_depth (int): The maximum depth of the tree.
                Will be updated for each level of nodes as the tree grows.
            stopping_criterion (callable): A function that takes the node/subtree as well as the data (`X`, `a`)
                and returns a boolean `True` if to stop splitting the tree and `False` if to continue splitting.
            max_splitting_values (int): The maximal number of unique values to consider when splitting a single feature
            multiple_hypothesis_test_method: The method for correcting p-values in multiple hypotheses testing.
                Should be compatible with statsmodels' `multipletests`.
            multiple_hypothesis_test_alpha (float): The alpha value for correcting p-values in
                multiple hypotheses testing.
                Should be compatible with statsmodels' `multipletests`.
            positivity_filtering_method (callable):
                A function that takes the current node (or subtree) as well as
                the arbitrary kwargs from `positivity_filtering_kwargs` and
                returns a list of the leaves/nodes' indices that do *not* violate positivity.
            positivity_filtering_kwargs: Keyword arguments to call the `positivity_filtering_method` with.
            _parent_ (BalancingTree): The parent of the current node/subtree.
        """
        self.asmd_violation_threshold = asmd_violation_threshold
        self.min_leaf_size = min_leaf_size
        self.min_treat_group_size = min_treat_group_size
        self.min_split_size = min_split_size
        self.max_depth = max_depth
        self.stopping_criterion = stopping_criterion
        self.max_splitting_values = max_splitting_values
        self.multiple_hypothesis_test_method = multiple_hypothesis_test_method
        self.multiple_hypothesis_test_alpha = multiple_hypothesis_test_alpha
        self.positivity_filtering_method = positivity_filtering_method
        self.positivity_filtering_kwargs = positivity_filtering_kwargs
        self._parent_ = _parent_

        if self.multiple_hypothesis_test_method is not None and self.multiple_hypothesis_test_alpha is None:
            raise ValueError(
                "Must be set `multiple_hypothesis_test_alpha` to a non-None value if "
                f"`multiple_hypothesis_test_method` (={multiple_hypothesis_test_method}) is provided."
            )

        # Add values that will be later initialized.from
        self.subtree_ = None
        self.node_index_ = None
        self.is_violating_positivity_ = None
        self.keep_ = False
        # TODO: I don't like the logic makes the default value of a newly generated node to "discard",
        #       Can I re-write the pruning logic that makes it possible to start with `keep_=True`?

    @property
    def _is_leaf(self):
        return self.subtree_ is None

    def fit(self, X: pd.DataFrame, a: pd.Series):
        """Builds a tree that stratifies the covariate space.

        Finds a stratification of the data space according to treatment allocation disparity by covariates in X
        The tree goes down until some stopping criteria is met, then prunes back according to the
        multiple hypothesis test results. Resulting leaf nodes are marked if their subpopulation violates positivity

        When the tree is pruned, nodes are kept if they have either or both:
            i) a descendant with a significant p-value
            ii) a direct parent with a significant p-value

        Args:
            X (pd.DataFrame): The feature matrix of all samples
            a (pd.Series): The treatment assignments

        Returns:
            BalancingTree:

        """
        self._build_tree(X, a)
        self._enumerate_nodes()
        if self.multiple_hypothesis_test_method is not None:
            self._prune(method=self.multiple_hypothesis_test_method, alpha=self.multiple_hypothesis_test_alpha)
        if self.positivity_filtering_method is not None:
            self._find_non_positivity_violating_leaves(
                positivity_filtering_method=self.positivity_filtering_method,
                positivity_filtering_kwargs=self.positivity_filtering_kwargs,
            )

        return self

    def _build_tree(self, X: pd.DataFrame, a: pd.Series):
        asmds = calculate_asmds(X, a)
        self.max_feature_asmd_ = asmds.max()
        self.treatment_prevalence_ = a.mean(axis=0)
        self.node_sample_size_ = len(a)
        if not self.stopping_criterion(self, X, a):
            self.split_feature_ = asmds.idxmax()
            self.split_value_ = self._find_split_value(X[self.split_feature_], a)
            if np.isnan(self.split_value_):
                self.split_feature_ = None
                return self
            self.n_asmd_violating_features_ = (asmds > self.asmd_violation_threshold).sum()
            self._recurse_over_build(X, a)

        # TODO: this building functionalities can be generalized, by generalizing the 3 main parts of this function:
        #       1. Feature selection: what feature is selected for splitting.
        #                             Inputs data (X, a, possibly y) and outputs a score per feature
        #                             (say, lower is better).
        #                             Currently ASMD, but with `y` may be Oui-ASMD, too, or something else.
        #                                   (However, will require `BalancingTree` to get optional `y`, too.)
        #       2. Feature splitting: Given a feature vector, how exactly to split it (to left and right children).
        #                             Inputs feature and `a`, and outputs a threshold.
        #                             Given how much of the pruning relies on p-values, it should also output a p-value
        #                             (or any other measure where score -> 0 is more confident and score -> 1 is less).
        #       3. Stopping criteria: Currently already implemented generically.
        #                             Inputs tree (node) and data (X, a, possibly y), and returns True if node should
        #                             not be further split.
        return self

    def _recurse_over_build(self, X, a):
        X_left, X_right, a_left, a_right = self._split_data(X, a)
        self_params = {v: getattr(self, v) for v in vars(self) if not v.endswith("_")}
        child_params = {
            **self_params,
            "max_depth": self_params["max_depth"] - 1,
            "_parent_": self,
        }
        self.subtree_ = (
            BalancingTree(**child_params)._build_tree(X_left, a_left),
            BalancingTree(**child_params)._build_tree(X_right, a_right),
        )

    def _split_data(self, X: pd.DataFrame, a: pd.Series = None):
        left = X[self.split_feature_] <= self.split_value_
        right = ~left  # note that this keeps NaNs on the right always
        res = []
        for data in [X, a]:
            if data is None:
                res.extend((None, None))
            else:
                res.extend((data[left], data[right]))
        return tuple(res)

    def _find_split_value(self, x: pd.Series, a: pd.Series):
        """Find a single split that should reduce imbalance
        by finding the value for which the mean treatment prevalence differs the most on both sides.

        Do not consider splits that result in groups that are smaller than min_leaf_size

        Args:
            x (pd.Series): The feature vector
            a (pd.Series): The treatment assignments

        Returns:
            The split value (normally a float, but could be anything that supports < comparison),
        """
        x_uniq = self._get_uniq_values(x)
        p_values = [self._fisher_test_p_value(x, a, x_val) for x_val in x_uniq]
        p_values = pd.Series(p_values, index=x_uniq).dropna()
        if len(p_values) == 0:
            return np.nan
        self.p_value_ = p_values.min()
        return p_values.idxmin()

    def _get_uniq_values(self, x):
        n = np.min([self.max_splitting_values, len(x)])
        x_uniq = x.quantile(q=np.arange(n + 1) / n, interpolation="lower").unique()
        x_uniq = x_uniq[x_uniq != x.max()]
        return x_uniq

    def _fisher_test_p_value(self, x, a, x_val):
        I = pd.cut(x, [-np.inf, x_val, np.inf], labels=[0, 1])
        if (sum(I == 0) <= self.min_leaf_size) | (sum(I == 1) <= self.min_leaf_size):
            return np.nan
        crosstab = pd.crosstab(I, a)
        if crosstab.min().min() < 20:
            _, p_value = fisher_exact(crosstab)
        else:
            _, p_value, _, _ = chi2_contingency(crosstab)
        return p_value

    def _enumerate_nodes(self):
        index = 0
        queue = deque([self])
        while queue:
            node = queue.popleft()
            node.node_index_ = index
            index += 1
            if not node._is_leaf:
                queue.extend(node.subtree_)

    def _prune(self, method: str, alpha: float):
        """Remove redundant nodes from tree.

        Redundant being their p-value does not pass multiple hypothesis correction

        1) implements the multiple hypothesis correction on all node p-values in the tree
        according to the user-defined method
        2) records the corrected p-value and test result for each node
        3) prunes back the tree, keeping only the nodes with:
            i) a descendant with a significant p-value
            ii) a direct parent with a significant p-value

        """
        # Consider using `alpha` and `method` from `self` rather than from parameters
        corrected_non_leaf_summary = self._multiple_hypothesis_correction(alpha, method)
        self._set_corrected_p_value_to_nodes(corrected_non_leaf_summary)
        self._recurse_over_prune()
        self._total_pruned_nodes_ = self._count_pruned_nodes()
        self._delete_post_pruning()

    def _multiple_hypothesis_correction(self, alpha: float, method: str):
        # take all non-leaf nodes p-values and conduct a multiple hypothesis correction
        non_leaf_summary = self._generate_non_leaf_nodes_summary().dropna(subset=["p_value"])
        if non_leaf_summary.empty:
            # Accounts for the edge case of a single-node tree
            test = (None, None)
        else:
            test = multipletests(non_leaf_summary['p_value'], alpha=alpha, method=method)

        # Record the corrected p-value and boolean result of the multiple hypothesis test:
        non_leaf_summary['corrected_pval'] = test[1]
        non_leaf_summary['multiple_test_result'] = test[0]
        return non_leaf_summary

    def _set_corrected_p_value_to_nodes(self, summary_df):
        if self._is_leaf:
            self.corrected_p_value_ = None
            self.corrected_p_value_is_significant_ = None
        else:
            self.corrected_p_value_ = summary_df['corrected_pval'].loc[
                summary_df['node_index'] == self.node_index_
            ].iloc[0]  # converting a single line Series into a float
            self.corrected_p_value_is_significant_ = (
                summary_df['multiple_test_result'].loc[summary_df['node_index'] == self.node_index_]
            ).iloc[0]
            self.subtree_[0]._set_corrected_p_value_to_nodes(summary_df)
            self.subtree_[1]._set_corrected_p_value_to_nodes(summary_df)

    def _recurse_over_prune(self):
        is_root = self._parent_ is None
        if is_root:
            self.keep_ = True
            pass
        if not self._is_leaf:
            if self.corrected_p_value_is_significant_:
                # the split is significant so we mark two children as keep
                # the ancestors should be marked as keep to eventually save this split
                self._mark_all_ancestors_and_two_children_as_keep()
            self.subtree_[0]._recurse_over_prune()
            self.subtree_[1]._recurse_over_prune()

    def _mark_all_ancestors_and_two_children_as_keep(self):
        self.keep_ = True
        self._mark_two_children_as_keep()
        is_root = self._parent_ is None
        if not is_root:
            self._parent_._mark_all_ancestors_and_two_children_as_keep()

    def _mark_two_children_as_keep(self):
        self.subtree_[0].keep_ = True
        self.subtree_[1].keep_ = True

    def _delete_post_pruning(self):
        if not self._is_leaf:
            # Two children will always have the same keep_ value
            assert self.subtree_[0].keep_ == self.subtree_[1].keep_
            if not self.subtree_[0].keep_:
                self.subtree_ = None
            else:
                self.subtree_[0]._delete_post_pruning()
                self.subtree_[1]._delete_post_pruning()

    def _count_pruned_nodes(self):
        count = 0
        queue = deque([self])
        while queue:
            node = queue.popleft()
            count += not node.keep_
            if not node._is_leaf:
                queue.extend(node.subtree_)
        return count

    def _generate_node_summary(self, X=None, a=None, y=None):
        # TODO: This function should be decomposed into two:
        #       1. `get_attributes(self)`, which gets the value of inherent node values:
        #          node_index, is_leaf, depth, positivity_violation, but also p_value (and corrected_p_value).
        #       2. `calculate_attributes(self, X, a)`, which gets data-dependent node values:
        #          treatment_prevalence, sample_size, max_asmd, and n_asmd_violating_features.
        #          These could be then calculated on new unseen data, rather than stored values seen during `fit`.
        node_summary = pd.DataFrame({
            "node_index": self.node_index_,
            "is_leaf": self._is_leaf,
            "max_depth": self.max_depth,  # Current node's depth is root.max_depth - node.max_depth
            "positivity_violation": self.is_violating_positivity_,
            # `getattr` allows to ask for a value before this post-fit attribute is set.
            # Ideally, this node summary function will work on an attribute-to-attribute requested basis,
            # and not return all possible attributes at once, avoiding the current requirement for
            # accessing attributes not yet set. but in the meantime:
            "p_value": getattr(self, "p_value_", None),
            "treatment_prevalence": getattr(self, "treatment_prevalence_", None),
            "sample_size": getattr(self, "node_sample_size_", None),
            "n_asmd_violating_features": getattr(self, "n_asmd_violating_features_", None),
            "max_asmd": getattr(self, "max_feature_asmd_", None),
        }, index=[0])
        if self._is_leaf:
            return node_summary
        else:
            left_child_summary = self.subtree_[0]._generate_node_summary()
            right_child_summary = self.subtree_[1]._generate_node_summary()
            return pd.concat([node_summary, left_child_summary, right_child_summary], axis=0, ignore_index=True)

    def generate_leaf_summary(self, X=None, a=None, y=None):
        """
        Recursively building a dataframe describing the leaf nodes with treatment_prevalence, sample_size,
        node_prevalence, node_index, average_outcome_treated/untreated, p_value, is_leaf, positivity_violation,
        max_depth, n_asmd_violating_features, max_asmd

        Args:
            X (pd.DataFrame): The feature data. Assumed to be of the same column
            structure as the training data
            a (pd.Series): The treatment assignment vector
            y (pd.Series): The outcome vector

        Returns: (DataFrame) a frame describing the leaf nodes in the fitted tree, with treatment_prevalence,
            sample_size, node_prevalence, node_index, average_outcome_treated/untreated, p_value, is_leaf,
            positivity_violation, max_depth, n_asmd_violating_features, max_asmd

        """
        all_node_summary = self._generate_node_summary(X, a, y)
        return all_node_summary.loc[all_node_summary["is_leaf"]]

    def _generate_non_leaf_nodes_summary(self, X=None, a=None, y=None):
        all_node_summary = self._generate_node_summary(X, a, y)
        return all_node_summary.loc[~all_node_summary["is_leaf"]]

    def _find_non_positivity_violating_leaves(
        self,
        positivity_filtering_method,
        positivity_filtering_kwargs: dict = None,
    ):
        """

        Finding the non-positivity-violating leaf nodes according to a user-defined criterion,
        marks the corresponding nodes, and outputs a list of non-violating node indices

        Args:
            positivity_filtering_method (function): user-defined function for determining positivity violations
            positivity_filtering_kwargs (dict): parameters for positivity criterion

        Returns: (list) node indices of leaves that do not violate positivity

        """
        non_violating_nodes_idx = positivity_filtering_method(self, **positivity_filtering_kwargs)  # returns the non-violating leaves
        self._mark_positivity_nodes(non_violating_nodes_idx)
        return non_violating_nodes_idx

    def _mark_positivity_nodes(self, non_positivity_violating_nodes):
        if not self._is_leaf:
            self.subtree_[0]._mark_positivity_nodes(non_positivity_violating_nodes)
            self.subtree_[1]._mark_positivity_nodes(non_positivity_violating_nodes)
        else:
            self.is_violating_positivity_ = self.node_index_ not in non_positivity_violating_nodes

    def get_positivity_violation_status(self) -> dict:
        """For each leaf node get whether it is violating positivity or not.

        Returns:
            dict: keys are node's index and value is boolean whether the leaf is violating.
        """
        if self._is_leaf:
            result = {self.node_index_: self.is_violating_positivity_}
        else:
            left = self.subtree_[0].get_positivity_violation_status()
            right = self.subtree_[1].get_positivity_violation_status()
            result = {**left, **right}
        return result

    def apply(self, X: pd.DataFrame):
        """
        Sends X down the fitted tree and outputs which leaf node each sample from X is assigned to

        Args:
            X: (pd.DataFrame): The feature matrix

        Returns: (pd.Series) leaf index associated with each sample indexed according to X

        """
        if self._is_leaf:
            return pd.Series(data=self.node_index_, index=X.index)  # pandas series of leaf, individual index

        else:
            X_left, X_right, *_ = self._split_data(X)
            left_child_assignment = self.subtree_[0].apply(X_left)
            right_child_assignment = self.subtree_[1].apply(X_right)
            node_assignment = pd.concat([left_child_assignment, right_child_assignment], axis="index")
            node_assignment = node_assignment.loc[X.index]  # Keeping order of X input
            return node_assignment

    def explain(
            self, X, a, split_condition: str = None, df_explanation: pd.DataFrame = None
    ):
        """Create a list of data frames summarizing the decision tree and the marginal effect.
        Each data-frame represents a leaf in the tree, and the list represents the tree itself.
        Each data frame exhibits several summary statistics about the path from the root to the
        leaf, including the maximal asmd at that level and the marginal outcome value.
        Args:
            X (pd.DataFrame): The feature data. Assumed to be of the same column
            structure as the training data
            a (pd.Series): The treatment assignment vector
            split_condition (str): The string representing the first condition. Default to 'All'
            df_explanation (pd.DataFrame): Mostly used for recursion. The initial data-frame,
            to which the split explanations are added
        Returns:
            List[pd.DataFrame]: The list representing the tree, holding a data-frame for every leaf.
        """
        line = self._create_df_line(X, a, split_condition)
        df_explanation = pd.DataFrame() if df_explanation is None else df_explanation
        df_explanation = pd.concat([df_explanation.T, line], axis=1).T
        df_explanation = self._convert_dtypes(df_explanation)
        if self._is_leaf:
            return [df_explanation]
        else:
            X_left, X_right, a_left, a_right = self._split_data(X, a)
            explanation_left = self._build_split_condition_string("<=")
            explanation_right = self._build_split_condition_string(">")
            return [
                self.subtree_[0].explain(X_left, a_left, explanation_left, df_explanation),
                self.subtree_[1].explain(X_right, a_right, explanation_right, df_explanation),
            ]

    def _build_split_condition_string(self, condition: str):
        """Create the condition string `feature <condition> value`"""
        explanation = f"{self.split_feature_} "
        explanation += condition
        explanation += f" {self.split_value_:.4g}"
        return explanation

    def _convert_dtypes(self, df_explanation: pd.DataFrame):
        df_explanation = df_explanation.astype(
            {
                "N": int,
                "Propensity": float,
                "asmd": float,
                "n_asmd_violating_features": int,
                # "effect": float,
            }
        )
        return df_explanation

    def _create_df_line(self, X, a, split_condition):
        split_condition = split_condition or "All"
        # marginal = self.outcome_learner.estimate_population_outcome(X, a, y)
        # effect = marginal.diff().values[-1]
        asmds = calculate_asmds(X, a)
        line = pd.Series(
            {
                "N": len(X.index),
                "Propensity": a.sum() / len(X.index),
                "asmd": asmds.max(),
                "n_asmd_violating_features": (asmds > self.asmd_violation_threshold).sum(),
                # "effect": effect,
            },
            name=split_condition,
        )
        return line



import unittest
from causallib.positivity import Trimming
from causallib.positivity.metrics import cross_covariance, cross_covariance_score
from causallib.positivity.datasets.positivity_data_simulator import make_multivariate_normal_data
from sklearn.linear_model import LogisticRegression
import pandas as pd
import numpy as np


class TestScoring(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        # Propensity model without regularization
        cls.trimming = Trimming(
            learner=LogisticRegression(solver="sag", penalty=None),
            threshold="crump"
        )
        # set mean and covariance params
        cls.treatment_2d_params = ([0, 0], [[2, -1], [-1, 2]])
        cls.control_2d_params = ([2, 2], [[1, 0], [0, 1]])
        cls.data_normal_2d = make_multivariate_normal_data(
            treatment_params=cls.treatment_2d_params,
            control_params=cls.control_2d_params)

    def _check_shape_of_cross_cov(self, X, cross_cov):
        for axis in cross_cov.shape:
            self.assertEqual(X.shape[1], axis)

    def _check_output_float_of_a_scorer(self, cross_cov_scorer):
        self.assertTrue(isinstance(cross_cov_scorer, float))

    def _check_output_list_of_a_scorer(self, cross_cov_scorer):
        self.assertTrue(isinstance(cross_cov_scorer, list))

    def test_cross_cov_shapes(self):
        X, a = self.data_normal_2d
        cross_cov_control = cross_covariance(X, a, mean_group=0)
        cross_cov_treatment = cross_covariance(X, a, mean_group=1)
        self._check_shape_of_cross_cov(X, cross_cov_control)
        self._check_shape_of_cross_cov(X, cross_cov_treatment)

    def test_raises_value_error_on_bad_value(self):
        X, a = self.data_normal_2d
        with self.assertRaises(ValueError):
            cross_covariance(X, a, mean_group=3)

    def test_cross_cov_scorer_reducing_functions_to_float(self):
        X, a = self.data_normal_2d
        cross_cov_max_score = cross_covariance_score(X, a, func=np.max)
        cross_cov_mean_trace_score = cross_covariance_score(
            X, a, func=lambda x: np.mean(np.diag(x)))
        self._check_output_float_of_a_scorer(cross_cov_max_score)
        self._check_output_float_of_a_scorer(cross_cov_mean_trace_score)

    def test_cross_cov_scorer_without_summing_scores(self):
        X, a = self.data_normal_2d
        cross_cov_scorer = cross_covariance_score(X, a, sum_scores=False)
        self._check_output_list_of_a_scorer(cross_cov_scorer)
        cross_cov_control_max = np.max(cross_covariance(X, a, mean_group=0))
        self.assertEqual(cross_cov_scorer[0], cross_cov_control_max)

    def test_cross_cov_scorer_off_diagonal(self):
        X, a = self.data_normal_2d
        cross_cov_off_diagonal = cross_covariance_score(
            X, a, off_diagonal_only=True, sum_scores=False, func=lambda x: x)
        self._check_output_list_of_a_scorer(cross_cov_off_diagonal)
        for i in range(len(cross_cov_off_diagonal)):
            self.assertEqual(np.trace(cross_cov_off_diagonal[i]), 0)
            self.assertEqual(np.argmax(cross_cov_off_diagonal[1], axis=0)[0], 1)

    def test_cross_cov_scorer_normalization(self):
        X, a = self.data_normal_2d
        cross_cov_score = cross_covariance_score(X, a)
        cross_cov_normalized_score = cross_covariance_score(
            X, a, normalize=True)
        self._check_output_float_of_a_scorer(cross_cov_score)
        self.assertTrue(cross_cov_normalized_score < cross_cov_score)

    def test_cross_cov_equiv_to_cov_if_done_on_the_same_group(self):
        """
        test that the cross-covariance is equal to covariance
        when applied on the same group
        """
        X, a = self.data_normal_2d
        X_duplicated_treated = pd.concat(
            [X.loc[a == 1, :], X.loc[a == 1, :]], axis=0).reset_index(drop=True)
        a = pd.Series(np.concatenate([np.ones_like(a.loc[a == 1]),
                                      np.zeros_like(a.loc[a == 1])]))
        cross_cov = cross_covariance(X_duplicated_treated, a)
        # true "regular" covariance matrix
        cov = np.cov(X_duplicated_treated.loc[a == 1, :], rowvar=False)

        self._check_shape_of_cross_cov(cov, cross_cov)
        np.testing.assert_array_almost_equal(cross_cov, cov, decimal=6)
        # test variance is close to 1 when normalizing similar groups
        # (not exactly 1 because only part of data is used for cross-cov)
        cross_cov_normalized = cross_covariance_score(
            X_duplicated_treated, a, normalize=True,
            func=lambda x: np.max(np.diag(x)), sum_scores=False)
        self.assertAlmostEqual(1, np.mean(cross_cov_normalized), 1)

    def test_trimming_cross_cov_score(self):
        """test scoring with trimming"""
        X, a = self.data_normal_2d
        self.trimming.fit(X, a)

        with self.subTest("Trimming score"):
            score_max = self.trimming.score(X, a)
            self._check_output_float_of_a_scorer(score_max)

        with self.subTest("Trimming score with mean function"):
            score_mean = self.trimming.score(X, a, func=np.mean)
            self.assertTrue(score_mean < score_max)

        with self.subTest("Trimming score without summing results"):
            score_mean_groups = self.trimming.score(X, a, func=np.mean,
                                                    sum_scores=False)
            self._check_output_list_of_a_scorer(score_mean_groups)
            self.assertTrue(np.max(score_mean_groups) < score_mean)

        with self.subTest("Trimming score off-diagonal"):
            score_max_off_diagonal = self.trimming.score(X, a,
                                                         off_diagonal_only=True)
            self.assertTrue(score_max_off_diagonal < score_max)

    def test_lower_score_with_trimming_in_non_overlap_case(self):
        X, a = self.data_normal_2d
        self.trimming.fit(X, a) # fit as default with crump method
        score_trimming = self.trimming.score(X, a)
        self.trimming.threshold_ = 0
        score_without_trimming = self.trimming.score(X, a)

        self.assertTrue(score_trimming < score_without_trimming)



from sklearn.linear_model import LogisticRegression, LinearRegression
from causallib.datasets.data_loader import load_nhefs_survival, load_nhefs
from causallib.estimation import IPW, Standardization
from causallib.estimation import Matching
from causallib.survival import WeightedSurvival
from causallib.evaluation import evaluate
import matplotlib.pyplot as plt

# data = load_nhefs_survival()
#
# matching = Matching(
#     with_replacement=True,
#     n_neighbors=1,
#     matching_mode="control_to_treatment",
# )
# surv_model = WeightedSurvival(matching)
# surv_model.fit(data.X, data.a, data.t, data.y)
# po = surv_model.estimate_population_outcome(data.X, data.a, data.t, data.y)
# print(po[1]-po[0])
#
# result = evaluate(matching, data.X, data.a, data.y)
# result.plot_all()

print("hey")

data = load_nhefs()

# ipw = IPW(LogisticRegression()).fit(data.X, data.a, data.y)
# result = evaluate(ipw, data.X, data.a, data.y, cv="auto")
# result.plot_all()
#
# std = Standardization(LinearRegression()).fit(data.X, data.a, data.y)
# result = evaluate(std, data.X, data.a, data.y)
# result.plot_all()
#
# matching = Matching(
#     with_replacement=True,
#     n_neighbors=1,
#     matching_mode="control_to_treatment",
# ).fit(data.X, data.a, data.y)
# result = evaluate(matching, data.X, data.a, data.y)
# result.plot_all()
#
# print("hey")

fig, axes = plt.subplots(1, 3, figsize=(9, 4))
ipw = IPW(LogisticRegression(max_iter=2)).fit(data.X, data.a, data.y)
result = evaluate(ipw, data.X, data.a, data.y)
result.plot_covariate_balance(kind="love", ax=axes[0])
result.plot_covariate_balance(kind="slope", thresh=0.3, ax=axes[1], plot_semi_grid=True, label_imbalanced=False)
result.plot_covariate_balance(kind="scatter", thresh=0.1, label_imbalanced=False, ax=axes[2])
print("hey")



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on Oct 26, 2021

"""

import unittest

import numpy as np
import pandas as pd
import math
from collections import deque

from causallib.datasets.data_loader import load_nhefs
from causallib.estimation import IPW, StratifiedStandardization
from sklearn.linear_model import LinearRegression, LogisticRegression

from causallib.contrib.bicause_tree import BICauseTree
from causallib.contrib.bicause_tree import PropensityBICauseTree
from causallib.contrib.bicause_tree.bicause_tree import BalancingTree
from causallib.contrib.bicause_tree.overlap_utils import prevalence_symmetric


class _BaseTestBICauseTree(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.a = pd.Series([0, 0, 0, 0, 0, 1, 1, 1, 1, 1] * 60, name="a")
        # constant feature - should never induce a split
        cls.feature0 = pd.Series([0, 0, 0, 0, 0, 0, 0, 0, 0, 0] * 60, name="x0")
        # biased features - should induce independent splits
        cls.feature1 = pd.Series([0, 0, 0, 0, 1, 1, 1, 1, 1, 0] * 60, name="x1")
        cls.feature2 = pd.Series([0, 0, 1, 1, 1, 1, 1, 1, 1, 0] * 60, name="x2")
        # outcomes with effect of 1
        cls.y = pd.Series([0, 0, 0, 0, 0, 1, 1, 1, 1, 1] * 60, name="y")
        cls.y_diverse = pd.Series([0, 0, 0, 0, 1, 2, 2, 2, 2, 1] * 60, name="yy")

    @staticmethod
    def load_nhefs_sample(n=50):
        nhefs = load_nhefs()
        X = nhefs.X.drop(columns=nhefs.X.filter(regex="education"))
        a = nhefs.a
        y = nhefs.y
        return X.iloc[:n], a.iloc[:n], y.iloc[:n]


class TestBICauseTreeDataStructure(_BaseTestBICauseTree):

    def test_large_two_level_input_splits(self):
        X = pd.DataFrame({"feature1": self.feature1})
        tree = BalancingTree(min_split_size=599).fit(X, self.a)
        self.assertEqual(tree.split_feature_, "feature1")
        self.assertEqual(tree.split_value_, 0)

    def test_two_splits(self):
        X = pd.DataFrame({"feature1": self.feature1, "feature2": self.feature2})
        tree = BalancingTree().fit(X, self.a)
        self.assertIsNotNone(tree.subtree_)
        self.assertIsInstance(tree.subtree_, tuple)
        self.assertEqual(len(tree.subtree_), 2)
        self.assertEqual(tree.subtree_[0].split_feature_, "feature2")
        self.assertFalse(hasattr(tree.subtree_[1], "split_feature_"))

    def test_split_size(self):
        X = pd.DataFrame({"feature1": self.feature1})
        tree = BalancingTree(min_split_size=600).fit(X, self.a)
        self.assertFalse(hasattr(tree, "split_feature_"))

    def test_stopping_max_depth(self):
        X = pd.DataFrame({"feature1": self.feature1, "feature2": self.feature2})
        tree = BalancingTree(max_depth=1).fit(X, self.a)
        self.assertIsNone(tree.subtree_)

    def test_stopping_min_split_size(self):
        X = pd.DataFrame({"feature1": self.feature1, "feature2": self.feature2})
        tree = BalancingTree(min_split_size=1200).fit(X, self.a)
        self.assertIsNone(tree.subtree_)

    def test_stopping_min_leaf_size(self):
        X = pd.DataFrame({"feature1": self.feature1, "feature2": self.feature2})
        tree = BalancingTree(min_leaf_size=1200).fit(X, self.a)
        self.assertIsNone(tree.subtree_)

    def test_stopping_min_treat_group_size(self):
        X = pd.DataFrame({"feature1": self.feature1, "feature2": self.feature2})
        tree = BalancingTree(min_treat_group_size=1200).fit(X, self.a)
        self.assertIsNone(tree.subtree_)

    def test_stopping_asmd_threshold(self):
        X = pd.DataFrame({"feature1": self.feature1, "feature2": self.feature2})
        tree = BalancingTree(asmd_violation_threshold=10 ** 4).fit(X, self.a)
        self.assertIsNone(tree.subtree_)

    def test_stopping_custom_function(self):
        def function_stop_immediately(tree, X, a):
            return True

        X = pd.DataFrame({"feature1": self.feature1, "feature2": self.feature2})
        tree = BalancingTree(stopping_criterion=function_stop_immediately).fit(X, self.a)
        self.assertIsNone(tree.subtree_)

    def test_prevalence_symmetric(self):
        X = pd.DataFrame({"feature1": self.feature1, "feature2": self.feature2})
        tree = BalancingTree().fit(X, self.a)
        node_idx = tree._find_non_positivity_violating_leaves(
            prevalence_symmetric,
            positivity_filtering_kwargs={'alpha': 0.5}
        )
        self.assertEqual(len(node_idx), 0)

    def test_multiple_hypothesis_bonferroni(self):
        tree = BalancingTree()
        tree.p_value_ = 0.01
        tree.subtree_ = []
        tree.subtree_.append(BalancingTree())
        tree.subtree_[0].p_value_ = 0.03
        tree.subtree_.append(BalancingTree())
        tree.subtree_[1].p_value_ = 0.02
        tree.subtree_[0].subtree_ = []
        tree.subtree_[0].subtree_.append(BalancingTree())
        tree.subtree_[0].subtree_[0].p_value_ = 0.04
        tree.subtree_[0].subtree_.append(BalancingTree())
        tree.subtree_[0].subtree_[1].p_value_ = 0.07
        tree._enumerate_nodes()
        set_node_attributes_helper(tree)
        corrected_non_leaf_summary = tree._multiple_hypothesis_correction(
            alpha=0.1, method='bonferroni'
        )
        tree._set_corrected_p_value_to_nodes(corrected_non_leaf_summary)
        self.assertEqual(tree.corrected_p_value_, 0.02)
        self.assertEqual(tree.subtree_[0].corrected_p_value_, 0.06)
        self.assertTrue(tree.corrected_p_value_is_significant_)
        self.assertTrue(tree.subtree_[0].corrected_p_value_is_significant_)

    def test_multiple_hypothesis_holm(self):
        tree = BalancingTree()
        tree.p_value_ = 0.01
        tree.subtree_ = []
        tree.subtree_.append(BalancingTree())
        tree.subtree_[0].p_value_ = 0.005
        tree.subtree_.append(BalancingTree())
        tree.subtree_[1].p_value_ = 0.02
        tree.subtree_[0].subtree_ = []
        tree.subtree_[0].subtree_.append(BalancingTree())
        tree.subtree_[0].subtree_[0].p_value_ = 0.03
        tree.subtree_[0].subtree_.append(BalancingTree())
        tree.subtree_[0].subtree_[1].p_value_ = 0.001
        tree.subtree_[1].subtree_ = []
        tree.subtree_[1].subtree_.append(BalancingTree())
        tree.subtree_[1].subtree_[0].p_value_ = 0.009
        tree.subtree_[1].subtree_.append(BalancingTree())
        tree.subtree_[1].subtree_[1].p_value_ = 0.018
        tree._enumerate_nodes()
        set_node_attributes_helper(tree)
        corrected_non_leaf_summary = tree._multiple_hypothesis_correction(
            alpha=0.1, method='holm'
        )
        tree._set_corrected_p_value_to_nodes(corrected_non_leaf_summary)
        self.assertEqual(tree.corrected_p_value_, 0.02)
        self.assertEqual(tree.subtree_[0].corrected_p_value_, 0.015)
        self.assertEqual(tree.subtree_[1].corrected_p_value_, 0.02)
        self.assertTrue(tree.corrected_p_value_is_significant_)
        self.assertTrue(tree.subtree_[0].corrected_p_value_is_significant_)
        self.assertTrue(tree.subtree_[1].corrected_p_value_is_significant_)

    def test_pruning_single_root_node(self):
        tree = BalancingTree()
        tree.corrected_p_value_is_significant_ = False
        tree._parent_ = None
        tree.subtree_ = None

        with self.subTest("Test pruning"):
            tree._recurse_over_prune()
            self.assertTrue(tree.keep_)

        with self.subTest("Test deletion after pruning"):
            tree._delete_post_pruning()
            self.assertIsNone(tree.subtree_)

    def test_no_pruning_needed(self):
        tree = BalancingTree()
        tree._parent_ = None
        tree.corrected_p_value_is_significant_ = False
        tree.subtree_ = []

        tree.subtree_.append(BalancingTree())
        tree.subtree_[0].corrected_p_value_is_significant_ = True
        tree.subtree_[0].subtree_ = []
        tree.subtree_[0].subtree_.append(BalancingTree())
        tree.subtree_[0].subtree_[0].corrected_p_value_is_significant_ = None
        tree.subtree_[0].subtree_[0].subtree_ = None
        tree.subtree_[0].subtree_.append(BalancingTree())
        tree.subtree_[0].subtree_[1].corrected_p_value_is_significant_ = None
        tree.subtree_[0].subtree_[1].subtree_ = None

        tree.subtree_.append(BalancingTree())
        tree.subtree_[1].corrected_p_value_is_significant_ = False
        tree.subtree_[1].subtree_ = []
        tree.subtree_[1].subtree_.append(BalancingTree())
        tree.subtree_[1].subtree_[0].corrected_p_value_is_significant_ = None
        tree.subtree_[1].subtree_[0].subtree_ = None
        tree.subtree_[1].subtree_.append(BalancingTree())
        tree.subtree_[1].subtree_[1].corrected_p_value_is_significant_ = None
        tree.subtree_[1].subtree_[1].subtree_ = None

        tree._enumerate_nodes()
        tree.subtree_[0]._parent_ = tree
        tree.subtree_[1]._parent_ = tree
        tree.subtree_[0].subtree_[0]._parent_ = tree.subtree_[0]
        tree.subtree_[0].subtree_[1]._parent_ = tree.subtree_[0]
        tree.subtree_[1].subtree_[0]._parent_ = tree.subtree_[1]
        tree.subtree_[1].subtree_[1]._parent_ = tree.subtree_[1]

        with self.subTest("Test pruning"):
            tree._recurse_over_prune()
            self.assertTrue(tree.keep_)
            self.assertTrue(tree.subtree_[0].keep_)
            self.assertTrue(tree.subtree_[0].subtree_[0].keep_)
            self.assertTrue(tree.subtree_[0].subtree_[1].keep_)
            self.assertTrue(tree.subtree_[1].keep_)

        with self.subTest("Test deletion after pruning"):
            tree._delete_post_pruning()
            self.assertIsNone(tree.subtree_[0].subtree_[0].subtree_)
            self.assertIsNone(tree.subtree_[0].subtree_[1].subtree_)
            self.assertIsNone(tree.subtree_[1].subtree_)

    def test_all_insignificant_p_values(self):
        tree = BalancingTree()
        tree._parent_ = None
        tree.corrected_p_value_is_significant_ = False
        tree.subtree_ = []

        tree.subtree_.append(BalancingTree())
        tree.subtree_[0].corrected_p_value_is_significant_ = False
        tree.subtree_[0].subtree_ = []
        tree.subtree_[0].subtree_.append(BalancingTree())
        tree.subtree_[0].subtree_[0].corrected_p_value_is_significant_ = None
        tree.subtree_[0].subtree_[0].subtree_ = None
        tree.subtree_[0].subtree_.append(BalancingTree())
        tree.subtree_[0].subtree_[1].corrected_p_value_is_significant_ = None
        tree.subtree_[0].subtree_[1].subtree_ = None

        tree.subtree_.append(BalancingTree())
        tree.subtree_[1].corrected_p_value_is_significant_ = False
        tree.subtree_[1].subtree_ = []
        tree.subtree_[1].subtree_.append(BalancingTree())
        tree.subtree_[1].subtree_[0].corrected_p_value_is_significant_ = None
        tree.subtree_[1].subtree_[0].subtree_ = None
        tree.subtree_[1].subtree_.append(BalancingTree())
        tree.subtree_[1].subtree_[1].corrected_p_value_is_significant_ = None
        tree.subtree_[1].subtree_[1].subtree_ = None

        tree._enumerate_nodes()
        tree.subtree_[0]._parent_ = tree
        tree.subtree_[1]._parent_ = tree
        tree.subtree_[0].subtree_[0]._parent_ = tree.subtree_[0]
        tree.subtree_[0].subtree_[1]._parent_ = tree.subtree_[0]
        tree.subtree_[1].subtree_[0]._parent_ = tree.subtree_[1]
        tree.subtree_[1].subtree_[1]._parent_ = tree.subtree_[1]

        with self.subTest("Test pruning"):
            tree._recurse_over_prune()
            self.assertTrue(tree.keep_)
            self.assertFalse(tree.subtree_[0].keep_)
            self.assertFalse(tree.subtree_[0].subtree_[0].keep_)
            self.assertFalse(tree.subtree_[0].subtree_[1].keep_)
            self.assertFalse(tree.subtree_[1].keep_)
            self.assertFalse(tree.subtree_[1].subtree_[0].keep_)
            self.assertFalse(tree.subtree_[1].subtree_[1].keep_)

        with self.subTest("Test deletion after pruning"):
            tree._delete_post_pruning()
            self.assertIsNone(tree.subtree_)

    def test_all_significant_p_values(self):
        tree = BalancingTree()
        tree._parent_ = None
        tree.corrected_p_value_is_significant_ = True
        tree.subtree_ = []

        tree.subtree_.append(BalancingTree())
        tree.subtree_[0].corrected_p_value_is_significant_ = True
        tree.subtree_[0].subtree_ = []
        tree.subtree_[0].subtree_.append(BalancingTree())
        tree.subtree_[0].subtree_[0].corrected_p_value_is_significant_ = None
        tree.subtree_[0].subtree_[0].subtree_ = None
        tree.subtree_[0].subtree_.append(BalancingTree())
        tree.subtree_[0].subtree_[1].corrected_p_value_is_significant_ = None
        tree.subtree_[0].subtree_[1].subtree_ = None

        tree.subtree_.append(BalancingTree())
        tree.subtree_[1].corrected_p_value_is_significant_ = True
        tree.subtree_[1].subtree_ = []
        tree.subtree_[1].subtree_.append(BalancingTree())
        tree.subtree_[1].subtree_[0].corrected_p_value_is_significant_ = None
        tree.subtree_[1].subtree_[0].subtree_ = None
        tree.subtree_[1].subtree_.append(BalancingTree())
        tree.subtree_[1].subtree_[1].corrected_p_value_is_significant_ = None
        tree.subtree_[1].subtree_[1].subtree_ = None

        tree._enumerate_nodes()
        tree.subtree_[0]._parent_ = tree
        tree.subtree_[1]._parent_ = tree
        tree.subtree_[0].subtree_[0]._parent_ = tree.subtree_[0]
        tree.subtree_[0].subtree_[1]._parent_ = tree.subtree_[0]
        tree.subtree_[1].subtree_[0]._parent_ = tree.subtree_[1]
        tree.subtree_[1].subtree_[1]._parent_ = tree.subtree_[1]

        with self.subTest("Test pruning"):
            tree._recurse_over_prune()
            self.assertTrue(tree.keep_)
            self.assertTrue(tree.subtree_[0].keep_)
            self.assertTrue(tree.subtree_[0].subtree_[0].keep_)
            self.assertTrue(tree.subtree_[0].subtree_[1].keep_)
            self.assertTrue(tree.subtree_[1].keep_)
            self.assertTrue(tree.subtree_[1].subtree_[0].keep_)
            self.assertTrue(tree.subtree_[1].subtree_[1].keep_)

        with self.subTest("Test deletion after pruning"):
            tree._delete_post_pruning()
            self.assertIsNone(tree.subtree_[0].subtree_[0].subtree_)
            self.assertIsNone(tree.subtree_[0].subtree_[1].subtree_)
            self.assertIsNone(tree.subtree_[1].subtree_[0].subtree_)
            self.assertIsNone(tree.subtree_[1].subtree_[1].subtree_)

    def test_mark_as_keep_and_prune1(self):
        tree = BalancingTree()
        tree._parent_ = None
        tree.corrected_p_value_is_significant_ = True
        tree.subtree_ = []

        tree.subtree_.append(BalancingTree())
        tree.subtree_[0].corrected_p_value_is_significant_ = False
        tree.subtree_[0].subtree_ = []
        tree.subtree_[0].subtree_.append(BalancingTree())
        tree.subtree_[0].subtree_[0].corrected_p_value_is_significant_ = None
        tree.subtree_[0].subtree_[0].subtree_ = None
        tree.subtree_[0].subtree_.append(BalancingTree())
        tree.subtree_[0].subtree_[1].corrected_p_value_is_significant_ = None
        tree.subtree_[0].subtree_[1].subtree_ = None

        tree.subtree_.append(BalancingTree())
        tree.subtree_[1].corrected_p_value_is_significant_ = False
        tree.subtree_[1].subtree_ = []
        tree.subtree_[1].subtree_.append(BalancingTree())
        tree.subtree_[1].subtree_[0].corrected_p_value_is_significant_ = None
        tree.subtree_[1].subtree_[0].subtree_ = None
        tree.subtree_[1].subtree_.append(BalancingTree())
        tree.subtree_[1].subtree_[1].corrected_p_value_is_significant_ = None
        tree.subtree_[1].subtree_[1].subtree_ = None

        tree._enumerate_nodes()
        tree.subtree_[0]._parent_ = tree
        tree.subtree_[1]._parent_ = tree
        tree.subtree_[0].subtree_[0]._parent_ = tree.subtree_[0]
        tree.subtree_[0].subtree_[1]._parent_ = tree.subtree_[0]
        tree.subtree_[1].subtree_[0]._parent_ = tree.subtree_[1]
        tree.subtree_[1].subtree_[1]._parent_ = tree.subtree_[1]

        with self.subTest("Test pruning"):
            tree._recurse_over_prune()
            self.assertTrue(tree.keep_)
            self.assertTrue(tree.subtree_[0].keep_)
            self.assertFalse(tree.subtree_[0].subtree_[0].keep_)
            self.assertFalse(tree.subtree_[0].subtree_[1].keep_)
            self.assertTrue(tree.subtree_[1].keep_)
            self.assertFalse(tree.subtree_[1].subtree_[0].keep_)
            self.assertFalse(tree.subtree_[1].subtree_[1].keep_)

        with self.subTest("Test deletion after pruning"):
            tree._delete_post_pruning()
            self.assertIsNone(tree.subtree_[0].subtree_)
            self.assertIsNone(tree.subtree_[1].subtree_)

    def test_mark_as_keep_and_prune2(self):
        tree = BalancingTree()
        tree.corrected_p_value_is_significant_ = False
        tree._parent_ = None
        tree.subtree_ = []

        tree.subtree_.append(BalancingTree())
        tree.subtree_[0].corrected_p_value_is_significant_ = False
        tree.subtree_.append(BalancingTree())
        tree.subtree_[1].corrected_p_value_is_significant_ = False

        tree.subtree_[0].subtree_ = []
        tree.subtree_[0].subtree_.append(BalancingTree())
        tree.subtree_[0].subtree_[0].corrected_p_value_is_significant_ = True

        tree.subtree_[0].subtree_[0].subtree_ = []
        tree.subtree_[0].subtree_[0].subtree_.append(BalancingTree())
        tree.subtree_[0].subtree_[0].subtree_[0].corrected_p_value_is_significant_ = None
        tree.subtree_[0].subtree_[0].subtree_[0].subtree_ = None
        tree.subtree_[0].subtree_[0].subtree_.append(BalancingTree())
        tree.subtree_[0].subtree_[0].subtree_[1].corrected_p_value_is_significant_ = None
        tree.subtree_[0].subtree_[0].subtree_[1].subtree_ = None

        tree.subtree_[0].subtree_.append(BalancingTree())
        tree.subtree_[0].subtree_[1].corrected_p_value_is_significant_ = None
        tree.subtree_[0].subtree_[1].subtree_ = None

        tree.subtree_[1].subtree_ = []
        tree.subtree_[1].subtree_.append(BalancingTree())
        tree.subtree_[1].subtree_[0].corrected_p_value_is_significant_ = None
        tree.subtree_[1].subtree_[0].subtree_ = None
        tree.subtree_[1].subtree_.append(BalancingTree())
        tree.subtree_[1].subtree_[1].corrected_p_value_is_significant_ = None
        tree.subtree_[1].subtree_[1].subtree_ = None

        tree._enumerate_nodes()
        tree.subtree_[0]._parent_ = tree
        tree.subtree_[1]._parent_ = tree
        tree.subtree_[0].subtree_[0]._parent_ = tree.subtree_[0]
        tree.subtree_[0].subtree_[1]._parent_ = tree.subtree_[0]
        tree.subtree_[1].subtree_[0]._parent_ = tree.subtree_[1]
        tree.subtree_[1].subtree_[1]._parent_ = tree.subtree_[1]

        with self.subTest("Test pruning"):
            tree._recurse_over_prune()
            self.assertTrue(tree.keep_)
            self.assertTrue(tree.subtree_[0].keep_)
            self.assertTrue(tree.subtree_[1].keep_)
            self.assertTrue(tree.subtree_[0].subtree_[0].keep_)
            self.assertTrue(tree.subtree_[0].subtree_[1].keep_)
            self.assertFalse(tree.subtree_[1].subtree_[0].keep_)
            self.assertFalse(tree.subtree_[1].subtree_[1].keep_)

        with self.subTest("Test deletion after pruning"):
            tree._delete_post_pruning()
            self.assertIsNone(tree.subtree_[0].subtree_[0].subtree_[0].subtree_)
            self.assertIsNone(tree.subtree_[0].subtree_[0].subtree_[1].subtree_)
            self.assertIsNone(tree.subtree_[0].subtree_[1].subtree_)
            self.assertIsNone(tree.subtree_[1].subtree_)


class TestBICauseTree(_BaseTestBICauseTree):
    def test_outcome_learners(self):
        from causallib.estimation import IPW, Standardization, MarginalOutcomeEstimator, TMLE, AIPW
        from sklearn.linear_model import LogisticRegression, LinearRegression
        import warnings
        from sklearn.exceptions import ConvergenceWarning

        X = pd.DataFrame({"feature1": self.feature1, "feature2": self.feature2})
        learners = [
            None,
            IPW(LogisticRegression()),
            Standardization(LinearRegression()),
            MarginalOutcomeEstimator(learner=None),
            # TMLE(Standardization(LinearRegression()), IPW(LogisticRegression())),
            AIPW(Standardization(LinearRegression()), IPW(LogisticRegression()))
        ]
        for learner in learners:
            learner_name = str(learner).split("(", maxsplit=1)[0]
            with self.subTest(f"Test fit using {learner_name}"), warnings.catch_warnings():
                warnings.simplefilter('ignore', category=ConvergenceWarning)
                tree = BICauseTree(learner)
                tree.fit(X, self.a, self.y)
                tree.estimate_population_outcome(X, self.a, self.y)
                if learner is None:  # To match expected default learner
                    learner = MarginalOutcomeEstimator(None)
                self.assertIsInstance(
                    list(tree.node_models_.values())[0], learner.__class__
                )

    def test_standardization_in_nodes_with_individual_prediction(self):
        # standardization + individual -> works
        X, a, y = self.load_nhefs_sample()

        tree = BICauseTree(
            outcome_model=StratifiedStandardization(LinearRegression()),
            individual=True
        )
        tree.fit(X, a, y)
        outcomes = tree.estimate_individual_outcome(X, a)  # y=None
        outcomes = tree.estimate_individual_outcome(X, a, y)
        leaf_count = len(tree.node_models_.keys())
        num_unique_predictions = outcomes.nunique(dropna=False)[0]
        self.assertGreater(num_unique_predictions, leaf_count)

    def test_standardization_in_nodes_with_population_prediction(self):
        # standardization + population -> works
        X, a, y = self.load_nhefs_sample()

        tree = BICauseTree(
            outcome_model=StratifiedStandardization(LinearRegression()),
            individual=False
        )
        tree.fit(X, a, y)
        outcomes = tree.estimate_individual_outcome(X, a)  # y=None
        outcomes = tree.estimate_individual_outcome(X, a, y)
        num_unique_predictions = outcomes.nunique(dropna=False)[0]
        leaf_count = len(tree.node_models_.keys())
        self.assertEqual(num_unique_predictions, leaf_count)

    def test_weight_model_in_nodes_with_population_prediction(self):
        # IPW + population -> works
        X, a, y = self.load_nhefs_sample()

        tree = BICauseTree(
            outcome_model=IPW(LogisticRegression()),
            individual=False
        )
        tree.fit(X, a, y)
        outcomes = tree.estimate_individual_outcome(X, a, y)
        # each node has the same values for each treatment group for all subjects.
        # so the number of values is the number of nodes times two.
        num_unique_predictions = outcomes.nunique(dropna=False)[0]
        leaf_count = len(tree.node_models_.keys())
        self.assertEqual(num_unique_predictions, leaf_count)

        with self.subTest("Not providing an outcome to pop-out-est raises an exception"):
            with self.assertRaises(TypeError):
                tree.estimate_individual_outcome(X, a, y=None)

    def test_weight_model_in_nodes_with_individual_prediction(self):
        # IPW + individual -> raise exception
        X, a, y = self.load_nhefs_sample()

        tree = BICauseTree(
            outcome_model=IPW(LogisticRegression()),
            individual=True
        )
        tree.fit(X, a, y)

        with self.assertRaises(Exception):
            tree.estimate_individual_outcome(X, a, y)

    def test_split_outcome(self):
        X = pd.DataFrame({"feature1": self.feature1})
        tree = BICauseTree(min_split_size=599).fit(
            X, self.a, self.y_diverse
        )

        self.assertListEqual(
            list(tree.estimate_population_outcome(X, self.a, self.y_diverse)),
            [0.5, 1.5],
        )

    def test_positivity_violation_creates_nans(self):
        X = pd.DataFrame({"featureA": self.a})
        tree = BICauseTree().fit(X, self.a, self.y)
        po = tree.estimate_population_outcome(X, self.a, self.y)

        self.assertSetEqual(set(po.index), set(self.a))
        np.testing.assert_array_equal(po, np.nan)
        np.testing.assert_equal(tree.tree.subtree_[0].max_feature_asmd_, np.nan)
        np.testing.assert_equal(tree.tree.subtree_[1].max_feature_asmd_, np.nan)

    def test_that_positivity_does_not_get_into_outcome(self):
        X = pd.DataFrame({"feature1": self.feature1, "feature2": self.feature2})
        tree = BICauseTree().fit(X, self.a, self.y)
        tree.fit_outcome_models(X, self.a, self.y)
        self.assertListEqual(
            list(tree.estimate_population_outcome(X, self.a, self.feature1)),
            [0.625, 0.625],
        )

    def test_large_constant_input_unchanged_and_returns_no_feature(self):
        X = pd.DataFrame({"feature0": self.feature0})
        tree = BICauseTree().fit(X, self.a, self.y)
        outcomes = tree.estimate_population_outcome(X, self.a, self.y)
        self.assertListEqual(list([outcomes[0], outcomes[1]]), [0, 1])
        self.assertFalse(hasattr(tree.tree, "split_feature_"))
        self.assertFalse(hasattr(tree.tree, "split_value_"))

    def test_parameter_propagation(self):
        from causallib.contrib.bicause_tree.overlap_utils import crump

        def function_stop_immediately(tree, X, a):
            return True
        # Define different-than-default values:
        kwargs = dict(
            min_leaf_size=5,
            min_split_size=3,
            min_treat_group_size=5,
            asmd_violation_threshold=0.3,
            max_depth=5,
            max_splitting_values=80,
            multiple_hypothesis_test_alpha=0.05,
            multiple_hypothesis_test_method='bonferroni',
            positivity_filtering_kwargs=dict(segments=500),
            stopping_criterion=function_stop_immediately,
            positivity_filtering_method=crump,
        )
        tree = BICauseTree(**kwargs)
        for k, v in kwargs.items():
            with self.subTest(f"Parameter propagation {k}"):
                tree_attr = getattr(tree.tree, k)
                self.assertEqual(v, tree_attr)


class TestPropensityBICauseTree(_BaseTestBICauseTree):
    def test_parameter_propagation(self):
        from causallib.contrib.bicause_tree.overlap_utils import crump

        def function_stop_immediately(tree, X, a):
            return True

        # Define different-than-default values:
        kwargs = dict(
            min_leaf_size=5,
            min_split_size=3,
            min_treat_group_size=5,
            asmd_violation_threshold=0.3,
            max_depth=5,
            max_splitting_values=80,
            multiple_hypothesis_test_alpha=0.05,
            multiple_hypothesis_test_method='bonferroni',
            positivity_filtering_kwargs=dict(segments=500),
            stopping_criterion=function_stop_immediately,
            positivity_filtering_method=crump,
        )
        tree = PropensityBICauseTree(**kwargs)
        for k, v in kwargs.items():
            with self.subTest(f"Parameter propagation {k}"):
                tree_attr = getattr(tree.tree, k)
                self.assertEqual(v, tree_attr)

    def test_prediction(self):
        X = pd.DataFrame({"feature1": self.feature1, "feature2": self.feature2})
        tree = PropensityBICauseTree().fit(X, self.a)
        a_pred = tree.predict_proba(X)
        # Compatible with sklearn's prediction:
        self.assertIsInstance(a_pred, np.ndarray)
        # Compatible with sklearn's `predict_proba` with column-per-class:
        self.assertTupleEqual(a_pred.shape, (X.shape[0], 2))
        # Result is a valid probability:
        self.assertLessEqual(a_pred.max(), 1)
        self.assertGreaterEqual(a_pred.min(), 0)

    def test_fit(self):
        from sklearn.dummy import DummyClassifier
        # Single level
        X = pd.DataFrame({"feature1": self.feature1})
        tree = PropensityBICauseTree().fit(X, self.a)
        self.assertSetEqual(set(tree.treatment_values_), set(self.a))
        self.assertEqual(len(tree.node_models_), 2)
        self.assertIsInstance(tree.node_models_[1], tree.learner.__class__)
        self.assertIsInstance(tree.node_models_[1], DummyClassifier)
        self.assertSetEqual(set(tree.node_models_[1].class_prior_), {0.2, 0.8})

    def test_non_default_learner(self):
        from sklearn.linear_model import LogisticRegression
        from sklearn.ensemble import GradientBoostingClassifier
        from sklearn.svm import LinearSVC

        X = pd.DataFrame({"feature1": self.feature1})
        learners = [
            LogisticRegression(), GradientBoostingClassifier(),
        ]
        for learner in learners:
            with self.subTest(f"Test `PropensityBICauseTree` with {learner}"):
                tree = PropensityBICauseTree(learner).fit(X, self.a)
                self.assertSetEqual(set(tree.treatment_values_), set(self.a))
                self.assertEqual(len(tree.node_models_), 2)
                self.assertIsInstance(tree.node_models_[1], tree.learner.__class__)
                self.assertIsInstance(tree.node_models_[1], learner.__class__)
                a_pred = tree.predict_proba(X)
                # Compatible with sklearn's prediction:
                self.assertIsInstance(a_pred, np.ndarray)
                # Compatible with sklearn's `predict_proba` with column-per-class:
                self.assertTupleEqual(a_pred.shape, (X.shape[0], 2))
                # Result is a valid probability:
                self.assertLessEqual(a_pred.max(), 1)
                self.assertGreaterEqual(a_pred.min(), 0)

        with self.subTest("Test with non-`predict_proba` learner"):
            tree = PropensityBICauseTree(LinearSVC())
            tree.fit(X, self.a)
            with self.assertRaises(AttributeError):
                tree.predict_proba(X)

    def test_fit_with_positivity_violation(self):
        from sklearn.dummy import DummyClassifier

        X = self.a.to_frame("x1")
        tree = PropensityBICauseTree().fit(X, self.a)
        a_pred = tree.predict_proba(X)
        self.assertEqual(len(tree.node_models_), 2)
        self.assertIsInstance(tree.node_models_[1], DummyClassifier)
        # Positivity violations result in extreme propensity scores:
        self.assertSetEqual(set(np.unique(a_pred)), {0, 1})
        # Verify column order so class 0 is 1st column:
        np.testing.assert_equal(a_pred[self.a == 0, 0], 1)
        np.testing.assert_equal(a_pred[self.a == 1, 1], 1)


def set_node_attributes_helper(node, default_val=0):
    # here we set the node attributes to a default value to be able to use the
    # _generate_node_summary method
    queue = deque([node])
    while queue:
        node = queue.popleft()
        node.node_sample_size_ = default_val
        node.treatment_prevalence_ = default_val
        node.potential_outcomes_ = pd.Series([default_val, default_val])
        if node.subtree_ is not None:
            queue.extend(node.subtree_)



"""
(C) Copyright 2019 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on Feb 21, 2019

"""

import unittest
import warnings

import pandas as pd

from causallib.utils import general_tools
from causallib.utils.stat_utils import robust_lookup
from causallib.utils.exceptions import ColumnNameChangeWarning


class TestUtils(unittest.TestCase):
    def ensure_learner_is_fitted(self, unfitted_learner, X, y):
        model_name = str(unfitted_learner.__class__).split(".")[-1]
        with self.subTest("Check is_fitted of {}".format(model_name)):
            self.assertFalse(general_tools.check_learner_is_fitted(unfitted_learner))
            unfitted_learner.fit(X, y)
            self.assertTrue(general_tools.check_learner_is_fitted(unfitted_learner))

    def test_check_classification_learner_is_fitted(self):
        from sklearn.linear_model import LogisticRegression
        from sklearn.tree import DecisionTreeClassifier
        from sklearn.ensemble import RandomForestClassifier
        from sklearn.svm import LinearSVC
        from sklearn.datasets import make_classification
        X, y = make_classification()
        for clf in [LogisticRegression(solver='lbfgs'), DecisionTreeClassifier(),
                    RandomForestClassifier(), LinearSVC()]:
            self.ensure_learner_is_fitted(clf, X, y)

    def test_check_regression_learner_is_fitted(self):
        from sklearn.linear_model import LinearRegression
        from sklearn.tree import ExtraTreeRegressor
        from sklearn.ensemble import GradientBoostingRegressor
        from sklearn.svm import SVR
        from sklearn.datasets import make_regression
        X, y = make_regression()
        for regr in [LinearRegression(), ExtraTreeRegressor(),
                     GradientBoostingRegressor(), SVR()]:
            self.ensure_learner_is_fitted(regr, X, y)

    def test_robust_lookup(self):
        import pandas as pd
        with self.subTest("zero-one columns, range index"):
            X = pd.DataFrame({
                0: [10, 20, 30, 40],
                1: [100, 200, 300, 400]
            })
            a = pd.Series([0, 1, 0, 1])
            extracted = robust_lookup(X, a)
            expected = pd.Series([10, 200, 30, 400])
            pd.testing.assert_series_equal(expected, extracted)

        with self.subTest("integer columns, range index"):
            X = pd.DataFrame({
                3: [10, 20, 30, 40],
                4: [100, 200, 300, 400]
            })
            a = pd.Series([3, 4, 3, 4])
            extracted = robust_lookup(X, a)
            expected = pd.Series([10, 200, 30, 400])
            pd.testing.assert_series_equal(expected, extracted)

        with self.subTest("text columns, range index"):
            X = pd.DataFrame({
                "a": [10, 20, 30, 40],
                "b": [100, 200, 300, 400]
            })
            a = pd.Series(["a", "b", "a", "b"])
            extracted = robust_lookup(X, a)
            expected = pd.Series([10, 200, 30, 400])
            pd.testing.assert_series_equal(expected, extracted)

        with self.subTest("text columns, text index"):
            X = pd.DataFrame({
                "a": [10, 20, 30, 40],
                "b": [100, 200, 300, 400]
            }, index=["w", "x", "y", "z"])
            a = pd.Series(["a", "b", "a", "b"], index=["w", "x", "y", "z"])
            extracted = robust_lookup(X, a)
            expected = pd.Series([10, 200, 30, 400], index=["w", "x", "y", "z"])
            pd.testing.assert_series_equal(expected, extracted)

        with self.subTest("index mismatch"):
            X = pd.DataFrame({
                0: [10, 20, 30, 40],
                1: [100, 200, 300, 400]
            }, index=[0, 1, 2, 3])
            a = pd.Series([0, 1, 0, 1], index=[0, 1, 2, 4])
            extracted = robust_lookup(X, a)
            expected = pd.Series([10, 200, 30, None], index=[0, 1, 2, 4])
            pd.testing.assert_series_equal(expected, extracted)

        with self.subTest("columns mismatch"):
            X = pd.DataFrame({
                0: [10, 20, 30, 40],
                1: [100, 200, 300, 400]
            })
            a = pd.Series([0, 1, 0, 2])
            extracted = robust_lookup(X, a)
            expected = pd.Series([10, 200, 30, None])
            pd.testing.assert_series_equal(expected, extracted)


class TestColumnNameSafeJoin(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.X = pd.DataFrame([[1, 10], [2, 20], [3, 30], [4, 40]])  # 4x2 dataframe
        cls.a = pd.Series([0, 0, 1, 1])
        cls.A = cls.X.copy()
        assert cls.X.shape[0] == cls.a.shape[0]

    # def setUp(self):
    #     # `ColumnNameChangeWarning` to act as an exception to be able to catch
    #     warnings.simplefilter("error", ColumnNameChangeWarning)

    def join_and_ensure_a_single_type(self, X, a):
        Xa = pd.concat([a, X], axis="columns")
        types = {type(col) for col in Xa.columns}
        self.assertEqual(1, len(types))

    def test_renaming_X_string_a_string(self):
        # `ColumnNameChangeWarning` to act as an exception to be able to catch
        warnings.simplefilter("error", ColumnNameChangeWarning)

        X = self.X.rename(columns={0: "x0", 1: "x1"})
        a = self.a.rename("a")
        # No `ColumnNameChangeWarning` should be raised
        reX, rea = general_tools.align_column_name_types_for_join(X, a)
        pd.testing.assert_frame_equal(X, reX)
        pd.testing.assert_series_equal(a, rea)
        self.join_and_ensure_a_single_type(reX, rea)

    def test_renaming_X_numeric_a_numeric(self):
        # `ColumnNameChangeWarning` to act as an exception to be able to catch
        warnings.simplefilter("error", ColumnNameChangeWarning)

        X = self.X.rename(columns={0: 0, 1: 1})
        a = self.a.rename(2)
        # No `ColumnNameChangeWarning` should be raised
        reX, rea = general_tools.align_column_name_types_for_join(X, a)
        pd.testing.assert_frame_equal(X, reX)
        pd.testing.assert_series_equal(a, rea)
        self.join_and_ensure_a_single_type(reX, rea)

    def test_renaming_X_string_a_numeric(self):
        X = self.X.rename(columns={0: "x0", 1: "x1"})
        a = self.a.rename(0)
        with self.assertWarns(ColumnNameChangeWarning):
            reX, rea = general_tools.align_column_name_types_for_join(X, a)
        pd.testing.assert_frame_equal(X, reX)
        self.assertEqual("0", rea.name)  # Stringify its name
        self.join_and_ensure_a_single_type(reX, rea)

    def test_renaming_X_numeric_a_string(self):
        X = self.X.rename(columns={0: 0, 1: 1})
        a = self.a.rename("a")
        with self.assertWarns(ColumnNameChangeWarning):
            reX, rea = general_tools.align_column_name_types_for_join(X, a)
        self.assertListEqual(["0", "1"], reX.columns.tolist())
        pd.testing.assert_series_equal(a, rea)
        self.join_and_ensure_a_single_type(reX, rea)

    def test_renaming_overwriting_a_name(self):
        X = self.X.rename(columns={0: "x0", 1: "x1"})
        a = self.a.rename("a")
        # with self.assertWarns(ColumnNameChangeWarning):
        reX, rea = general_tools.align_column_name_types_for_join(X, a, a_name="w")
        pd.testing.assert_frame_equal(X, reX)
        self.assertEqual("w", rea.name)  # Stringify its name
        self.join_and_ensure_a_single_type(reX, rea)

    def test_renaming_no_a_name(self):
        X = self.X.rename(columns={0: "x0", 1: "x1"})
        a = self.a.rename(None)
        with self.assertWarns(ColumnNameChangeWarning):
            reX, rea = general_tools.align_column_name_types_for_join(X, a)
        pd.testing.assert_frame_equal(X, reX)
        self.assertEqual("a", rea.name)  # Default value
        self.join_and_ensure_a_single_type(reX, rea)

    def test_renaming_X_string_A_numeric(self):
        X = self.X.rename(columns={0: "x0", 1: "x1"})
        A = self.A.copy()
        with self.assertWarns(ColumnNameChangeWarning):
            reX, reA = general_tools.align_column_name_types_for_join(X, A)
        pd.testing.assert_frame_equal(X, reX)
        self.assertEqual(["a_0", "a_1"], reA.columns.tolist())
        self.join_and_ensure_a_single_type(reX, reA)

    def test_renaming_X_numeric_A_string(self):
        X = self.X.rename(columns={0: 0, 1: 1})
        A = self.A.rename(columns={0: "a0", 1: "a1"})
        with self.assertWarns(ColumnNameChangeWarning):
            reX, reA = general_tools.align_column_name_types_for_join(X, A)
        self.assertListEqual(["0", "1"], reX.columns.tolist())
        self.assertListEqual(["a_a0", "a_a1"], reA.columns.tolist())
        self.join_and_ensure_a_single_type(reX, reA)

    def test_renaming_X_mixed_a_string(self):
        X = self.X.rename(columns={0: "x0", 1: 1})
        a = self.a.rename("a")
        with self.assertWarns(ColumnNameChangeWarning):
            reX, rea = general_tools.align_column_name_types_for_join(X, a)
        self.assertListEqual(["x0", "1"], reX.columns.tolist())
        pd.testing.assert_series_equal(a, rea)
        self.join_and_ensure_a_single_type(reX, rea)

    def test_renaming_X_mixed_a_numeric(self):
        X = self.X.rename(columns={0: "x0", 1: 1})
        a = self.a.rename(2)
        with self.assertWarns(ColumnNameChangeWarning):
            reX, rea = general_tools.align_column_name_types_for_join(X, a)
        self.assertListEqual(["x0", "1"], reX.columns.tolist())
        self.assertEqual("2", rea.name)  # Stringify its name
        self.join_and_ensure_a_single_type(reX, rea)

    # # ##Scikit-learn doesn't seem to care about duplicated names in `X` ## # #
    # def test_renaming_X_numeric_a_numeric_duplicated(self):
    #     X = self.X.rename(columns={0: 0, 1: 1})
    #     a = self.a.rename(0)
    #     # with self.assertWarns(ColumnNameChangeWarning):
    #     reX, rea = general_tools.align_column_name_types_for_join(X, a)
    #     pd.testing.assert_frame_equal(X, reX)
    #     self.assertEqual(2, rea.name)  # max(cols)=1 + 1 = 2
    #     self.join_and_ensure_a_single_type(reX, rea)

    def test_join_X_full_a_full(self):
        X = self.X.rename(columns={0: "x0", 1: "x1"})
        a = self.a.rename(None)
        Xa = general_tools.column_name_type_safe_join(X, a)

        with self.subTest("Column types"):
            types = {type(col) for col in Xa.columns}
            self.assertSetEqual({str}, types)

        with self.subTest("Content"):
            self.assertListEqual(a.tolist(), Xa.iloc[:, 0].tolist())
            pd.testing.assert_frame_equal(X, Xa.iloc[:, 1:])

    def test_join_X_empty_a_full(self):
        X = pd.DataFrame(index=self.a.index)
        a = self.a.rename("a")
        Xa = general_tools.column_name_type_safe_join(X, a)

        with self.subTest("Column types"):
            types = {type(col) for col in Xa.columns}
            self.assertSetEqual({str}, types)

        with self.subTest("Content"):
            self.assertEqual(1, Xa.shape[1])
            self.assertListEqual(a.tolist(), Xa.squeeze().tolist())

    def test_join_X_full_a_empty(self):
        X = self.X.copy()
        a = pd.Series(dtype=object)
        Xa = general_tools.column_name_type_safe_join(X, a)

        with self.subTest("Column types"):
            types = {type(col) for col in Xa.columns}
            self.assertSetEqual({int}, types)

        with self.subTest("Content"):
            self.assertEqual(2, Xa.shape[1])
            pd.testing.assert_frame_equal(X, Xa)

    def test_join_X_full_A_full(self):
        X = self.X.rename(columns={0: "x0", 1: "x1"})
        A = self.A.copy()
        Xa = general_tools.column_name_type_safe_join(X, A)

        with self.subTest("Column types"):
            types = {type(col) for col in Xa.columns}
            self.assertSetEqual({str}, types)

        with self.subTest("Content"):
            A.columns = ["_0", "_1"]
            pd.testing.assert_frame_equal(A, Xa.iloc[:, :2])
            pd.testing.assert_frame_equal(X, Xa.iloc[:, 2:])

    def test_join_X_empty_A_full(self):
        X = pd.DataFrame()
        A = self.A.copy()
        Xa = general_tools.column_name_type_safe_join(X, A)

        with self.subTest("Column types"):
            types = {type(col) for col in Xa.columns}
            self.assertSetEqual({int}, types)

        with self.subTest("Content"):
            self.assertEqual(2, Xa.shape[1])
            pd.testing.assert_frame_equal(A, Xa)

    def test_join_X_full_A_empty(self):
        X = self.X.rename(columns={0: "x0", 1: "x1"})
        A = pd.DataFrame()
        Xa = general_tools.column_name_type_safe_join(X, A)

        with self.subTest("Column types"):
            types = {type(col) for col in Xa.columns}
            self.assertSetEqual({str}, types)

        with self.subTest("Content"):
            self.assertEqual(2, Xa.shape[1])
            pd.testing.assert_frame_equal(X, Xa)



import numpy as np
import pandas as pd

from causallib.estimation.base_estimator import IndividualOutcomeEstimator


class OverlapViolationEstimator(IndividualOutcomeEstimator):
    """
    A causallib-compatible model returning NaNs with appropriate format and indexing.
    Meant to be used when overlap violations are detected.
    """
    def __init__(self, value=np.nan, *args, **kwargs) -> None:
        super().__init__(None, *args, **kwargs)
        self.value = value

    def fit(self, X, a, y, sample_weight=None):
        return self

    def estimate_population_outcome(self, X, a, y=None, **kwargs):
        treatment_values = self._get_treatment_index(a)
        return pd.Series(self.value, index=treatment_values)

    def estimate_individual_outcome(self, X, a, **kwargs):
        treatment_values = self._get_treatment_index(a)
        return pd.DataFrame(self.value, index=X.index, columns=treatment_values)

    @staticmethod
    def _get_treatment_index(a):
        return pd.Index(sorted(a.unique()), name=a.name)


def prevalence_symmetric(tree, alpha=0.1):

    """
    Computes the symmetric lower/upper cutoff based on prevalence and a fixed cutoff alpha
    with p the original propensity score and mu the prevalence in the cohort
    Returns the list of non violating nodes for these cutoffs

    Args:
        tree (BalancingTree): the splitting tree
        alpha (float): the fixed cutoff to be transformed

    Checks for the default stopping criterion, which is:
    is the ASMD threshold reached
    OR is the minimum treatment group size reached
    OR is the minimum leaf size reached
    OR is the maximum depth of the tree reached
    Returns: (list) The node indices of non-violating leaves
    """
    leaf_summary = tree.generate_leaf_summary()
    ps_singles_list = leaf_summary['treatment_prevalence'].values
    repetition_numbers = leaf_summary['sample_size'].values.tolist()
    ps_list = ps_singles_list.repeat(repetition_numbers)
    ps_vect = pd.DataFrame(np.column_stack((1 - ps_list, ps_list)))
    mu = tree.treatment_prevalence_  # Overall prevalence of treatment
    cutoffs = prevalence_symmetric_cutoff(ps_vect, mu, alpha)
    lower_cutoff, upper_cutoff = cutoffs[0], cutoffs[1]
    non_violating_nodes = leaf_summary.loc[
        leaf_summary['treatment_prevalence'].between(lower_cutoff, upper_cutoff),
        'node_index'
    ].tolist()
    return non_violating_nodes


def prevalence_symmetric_cutoff(prob, mu, alpha=0.1):
    """
    Computes a lower/upper cutoff based on the prevalence of
    treatment in the cohort. Treatment should be binary.

    Args:
        mu (float): observed prevalence of treatment in the cohort
        prob (pd.DataFrame): probability to be assigned to a group
                          (n_samples, 2)
                          For binary treatment each row is (1-p, p)
        alpha (float): the fixed cutoff to be transformed, should be
        strictly between 0 and 1

    Returns: Tuple[float, float]: upper and lower cutoff
    """
    if prob.shape[1] > 2:
        raise ValueError('This threshold selection method is applicable only '
                         'for binary treatment assignment')
    if not 0 < alpha < 1:
        raise ValueError(f"`alpha` value should be in the open interval (0, 1). Got {alpha} instead.")
    upper_cutoff = (1 - alpha)*mu / ((1-alpha)*mu + alpha*(1 - mu))
    lower_cutoff = alpha*mu / (alpha*mu + (1 - alpha)*(1 - mu))
    return lower_cutoff, upper_cutoff


def crump(tree, segments=10000):
    """
    Generates a list of non positivity violating node indices according to the crump cutoff
    i.e. nodes with propensity scores in [cutoff, 1-cutoff]

    Crump is a systematic approach to find the optimal trimming cutoff see
    "Crump, R. K., Hotz, V. J., Imbens, G. W., & Mitnik, O. A. (2009).
    Dealing with limited overlap in estimation of average treatment effects."

    Args:
        tree (class): the splitting tree
        segments (int): number of portions used for computing the crump cutoff


    Returns: (list) The node indices of non-violating leaves
    """
    # TODO: add data to signature once we remove the saved data attributes from leaves.

    leaf_summary = tree.generate_leaf_summary()
    ps_singles_list = leaf_summary['treatment_prevalence'].values
    repetition_numbers = leaf_summary['sample_size'].values.tolist()
    ps_list = ps_singles_list.repeat(repetition_numbers)
    ps_vect = pd.DataFrame(np.column_stack((1 - ps_list, ps_list)))
    lower_cutoff, upper_cutoff = crump_cutoff(ps_vect, segments)
    if lower_cutoff is not None:
        non_violating_nodes = leaf_summary.loc[
            leaf_summary['treatment_prevalence'].between(lower_cutoff, upper_cutoff),
            'node_index'
        ].tolist()
        return non_violating_nodes
    return []  # lower_cutoff=None means all nodes have positivity violations


def crump_cutoff(prob, segments=10000 ):
    """
    A systematic approach to find the optimal trimming cutoff, based on the
    marginal distribution of the propensity score,
    and according to a variance minimization criterion.
    "Crump, R. K., Hotz, V. J., Imbens, G. W., & Mitnik, O. A. (2009).
    Dealing with limited overlap in estimation of average treatment effects."
    Treatment should be binary.
    Args:
        prob (pd.DataFrame): probability to be assigned to a group
                          (n_samples, 2)
        segments (int): number of exclusive segments of the interval (0, 0.5].
                        more segments results with more precise cutoff
    Returns:
        float: the optimal cutoff,
               i.e. the smallest value that satisfies the criterion.
    """
    # TODO: rethink input - probability_matrix, propensity_vector, or model + data
    if prob.shape[1] > 2:
        raise ValueError('This threshold selection method is applicable only '
                         'for binary treatment assignment')
    else:
        propensities = prob.iloc[:, 1]

    alphas = np.linspace(1e-7, 0.5, segments)
    alphas_weights = alphas * (1 - alphas)
    overlap_weights = propensities * (1 - propensities)
    for i in range(segments):
        obs_meets_criterion = overlap_weights >= alphas_weights[i]
        criterion = 2 * (np.sum(obs_meets_criterion / overlap_weights) /
                         np.maximum(np.sum(obs_meets_criterion), 1e-7))
        if (1 / alphas_weights[i]) <= criterion:
            return alphas[i], 1-alphas[i]
    return None, None  # No overlap



import pandas as pd
import numpy as np
from causallib.positivity import BasePositivity
from sklearn.linear_model import LogisticRegression

OPTIMAL_THRESHOLD_ACCURACY = 5e-6


def _check_is_valid_threshold_value(threshold_value):
    if not isinstance(threshold_value, (float, type(None))):
        raise ValueError("invalid threshold_value")
    return threshold_value


def _check_is_valid_threshold_method(threshold_method):
    threshold_method = "crump" if threshold_method == "auto" else threshold_method
    if threshold_method not in cutoff_optimizers:
        raise ValueError("invalid threshold_method")
    return threshold_method


def _check_propensities(prob):
    """ check if the treatment assignment is binary"""
    if prob.shape[1] > 2:
        raise ValueError('This threshold selection method is applicable only '
                         'for binary treatment assignment')
    else:
        propensities = prob.iloc[:, 1]
        return propensities


def crump_cutoff(prob, segments=10000):
    """
    A systematic approach to find the optimal trimming cutoff, based on the
    marginal distribution of the propensity score,
    and according to a variance minimization criterion.

    "Crump, R. K., Hotz, V. J., Imbens, G. W., & Mitnik, O. A. (2009).
    Dealing with limited overlap in estimation of average treatment effects."
    Args:
        prob (pd.Series): probability of be assign to a group
                          (n_samples, n_classes)
        segments (int): number of exclusive segments of the interval (0, 0.5].
                        more segments results with more precise cutoff

    Returns:
        float: the optimal cutoff,
               i.e. the smallest value that satisfies the criterion.
    """
    propensities = _check_propensities(prob)
    alphas = np.linspace(1e-7, 0.5, segments)
    alphas_weights = alphas * (1 - alphas)
    overlap_weights = propensities * (1 - propensities)
    for i in range(segments):
        obs_meets_criterion = overlap_weights >= alphas_weights[i]
        criterion = 2 * (np.sum(obs_meets_criterion / overlap_weights) /
                         np.maximum(np.sum(obs_meets_criterion), 1e-7))
        if (1 / alphas_weights[i]) <= criterion:
            break
    return alphas[i]


cutoff_optimizers = {'crump': crump_cutoff}


def _lookup_method(threshold_method):
    if threshold_method in cutoff_optimizers:
        return cutoff_optimizers[threshold_method]
    else:
        raise Exception("Method %s does not exist" % threshold_method)


class Trimming(BasePositivity):
    def __init__(self,
                 learner=LogisticRegression(),
                 threshold="auto"):
        """

        Args:
            learner (sklearn object): Initialized sklearn model
            threshold (str | float) : The threshold method or value.
                - if auto: finding the optimized threshold in a principled way.
                - if float, hard-coded value between 0 to 0.5 is used
                  in order to clip the propensity estimation.
        """
        self.learner = learner
        if not hasattr(self.learner, "predict_proba"):
            raise AttributeError("Propensity Estimator must use a machine "
                                 "learning that can predict probabilities"
                                 "(i.e., have predict_proba method)")

        if isinstance(threshold, str):
            self.threshold = _check_is_valid_threshold_method(threshold)
        else:
            self.threshold_ = _check_is_valid_threshold_value(threshold)

    def _fit_threshold(self, X):
        """Fit threshold in a principled way"""
        prob = self.learner.predict_proba(X)
        prob = pd.DataFrame(prob, index=X.index, columns=self.learner.classes_)
        method = _lookup_method(self.threshold)
        threshold = method(prob)
        return threshold

    def fit(self, X, a):
        """Fit propensity model for positivity.

        Args:
            X (pd.DataFrame): covariate matrix of size
                              (num_subjects, num_features)
            a (pd.Series): treatment assignment of size (num_subjects,)
        """
        self.learner.fit(X, a)
        if hasattr(self, 'threshold'):
            self.threshold_ = self._fit_threshold(X)
        return self

    def predict(self, X, a, threshold=None):
        """Predict whether or not a sample is in the overlap region.
        Find samples that have probabilities to be assigned to one of the
        treatment groups, that is bigger than the cutoff threshold.

        return a boolean indexer which is `True` if their probabilities are
        higher than the cutoff threshold and `False` otherwise.
        Args:
            X (pd.DataFrame): covariate matrix of size
                              (num_subjects, num_features)
            a (pd.Series): treatment assignment of size (num_subjects,)
            threshold (float|None): The cutoff threshold.
                - if float, an optional value between 0 to 0.5 to clip the
                    propensity estimation.
                - if None, use the optimized cutoff in a principled way.

        Returns:
            pd.Series: a Series of length `X.shape[0]` with the same index as
               `X` and only boolean values
        """
        prob = self.learner.predict_proba(X)
        prob = pd.DataFrame(prob, index=X.index, columns=self.learner.classes_)

        threshold_value = _check_is_valid_threshold_value(threshold)
        threshold_to_use = (self.threshold_ if threshold_value is None
                            else threshold_value)

        untrimmed_indices = (prob >= threshold_to_use).all(axis=1)
        return untrimmed_indices



"""
(C) Copyright 2021 IBM Corp.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
Created on March 2, 2021
"""
from __future__ import annotations
from abc import ABC, abstractmethod
import pandas as pd
from typing import Tuple
from sklearn.base import BaseEstimator


class BasePositivity(ABC, BaseEstimator):

    @abstractmethod
    def fit(self,
            X: pd.DataFrame, a: pd.Series) -> BasePositivity:
        """Fit positivity checker.

        Args:
            X (pd.DataFrame): Covariate matrix of size (num_subjects, num_features).
            a (pd.Series): Treatment assignment of size (num_subjects,).
        """
        raise NotImplementedError

    @abstractmethod
    def predict(self, X: pd.DataFrame, a: pd.Series) -> pd.Series:
        """Predict whether a sample is in the overlap of treatments.

        Args:
            X (pd.DataFrame): Covariate matrix of size (num_subjects, num_features)
            a (pd.Series): Treatment assignment of size (num_subjects,).

        Returns:
            pd.Series: a Series of length `X.shape[0]` with the same index as
               `X` and only boolean values
        """
        raise NotImplementedError

    def transform(self,
                  X: pd.DataFrame, a: pd.Series, *args: pd.Series
                  ) -> Tuple[pd.DataFrame, pd.Series, pd.Series]:
        """Transform the input data to remove positivity violations.

        Args:
            X (pd.DataFrame): Covariate matrix of size (num_subjects, num_features).
            a (pd.Series): Treatment assignment of size (num_subjects,).
            *args (pd.Series): Zero or more pd.Series objects corresponding to 
               outcomes. Each argument must be indexed the same as the other 
               arguments and have size (num_subjects,).

        Returns:
            Tuple[pd.DataFrame, pd.Series, pd.Series]: Subsets of `X`, `a` and
            the output series objects of `args` corresponding to the samples
            which do not violate the positivity assumption.
        """
        indices_to_keep = self.predict(X, a)
        return_list = [X.loc[indices_to_keep], a.loc[indices_to_keep]]
        for output in args:
            return_list.append(output.loc[indices_to_keep])
        return return_list

    def fit_predict(self, X: pd.DataFrame, a: pd.Series) -> pd.Series:
        """Fit positivity checker and predict overlap membership.

        This is a convenience function that calls `fit` and `predict`.

        Args:
            X (pd.DataFrame): Covariate matrix of size (num_subjects, num_features)
            a (pd.Series): Treatment assignment of size (num_subjects,).

        Returns:
            pd.Series: a Series of length `X.shape[0]` with the same index as
               `X` and only boolean values
        """
        self.fit(X, a)
        return self.predict(X, a)

    def fit_transform(self,
                      X: pd.DataFrame, a: pd.Series, *args: pd.Series
                      ) -> Tuple[pd.DataFrame, pd.Series, pd.Series]:
        """Fit and transform data by removing positivity violations.

        Args:
            X (pd.DataFrame): Covariate matrix of size (num_subjects, num_features).
            a (pd.Series): Treatment assignment of size (num_subjects,).
            *args (pd.Series): Zero or more pd.Series objects corresponding to 
               outcomes. Each argument must be indexed the same as the other 
               arguments and have size (num_subjects,).

        Returns:
            Tuple[pd.DataFrame, pd.Series, pd.Series]: Subsets of `X`, `a` and
            the output series objects of `args` corresponding to the samples
            which do not violate the positivity assumption.
        """
        self.fit(X, a)
        return self.transform(X, a, *args)

    def score(self,
              X: pd.DataFrame, a: pd.Series,
              **kwargs):
        """Score the positivity violation
        This is a generic function, but right now it receives
        only one kind of scorer - cross_covaraince_score
        Args:
            X (pd.DataFrame): Covariate matrix of size (num_subjects, num_features).
            a (pd.Series): Treatment assignment of size (num_subjects,).
            **kwargs : kwargs that are corresponding to the scoring metric.

        Returns:
            float: a non-negative score that quantifies the violation
            of positivity
        """
        from .metrics.metrics import cross_covariance_score
        X_trans, a_trans = self.transform(X, a)
        return cross_covariance_score(X_trans, a_trans, **kwargs)




from causallib.metrics.scorers import PropensityScorerBase


class SKLearnScorerWrapper(PropensityScorerBase):
    def __init__(self, score_func, sign=None, **kwargs):
        super().__init__(
            score_func=score_func,
            sign=1,  # This keeps original scorer sign
            **kwargs
        )

    def _score(self, estimator, X, a, y=None, sample_weight=None, **kwargs):
        learner = self._extract_sklearn_estimator(estimator)
        score = self._score_func(learner, X, a, sample_weight=sample_weight)
        return score

    @staticmethod
    def _extract_sklearn_estimator(estimator):
        if hasattr(estimator, "best_estimator_"):
            # Causallib's wrapper around GridSearchCV
            return estimator.best_estimator_.learner
        if hasattr(estimator, "learner"):
            return estimator.learner
        raise AttributeError(
            f"Could not extract an sklearn estimator from {estimator},"
            f"which has the following attributes:\n"
            f"{list(estimator.__dict__.keys())}"
        )


"""
Metrics to asses violations of positivity
"""

import numpy as np
import pandas as pd


def _check_mean_group(mean_group):
    mean_group = int(mean_group)
    if mean_group not in {0, 1}:
        raise ValueError('mean_group needs to be equal to 0 or 1')
    return mean_group


def cross_covariance(X, a, mean_group=0):
    """
    Computing the covariance, where the mean is taken from the counter group.
    suitable only for binary treatment, i.e a \in {0, 1}
    Args:
        X (pd.DataFrame): Covariate matrix of size (num_subjects, num_features).
        a (pd.Series): Treatment assignment of size (num_subjects,).
        mean_group (0 or 1): the treatment group

    Returns:
        np.array: cross-covariance square matrix
    """
    mean_group = _check_mean_group(mean_group)
    avg = X.loc[a == mean_group, :].mean(axis=0)
    X_counter_group = X.loc[a == int(not mean_group)]
    X_counter_group = (X_counter_group - avg).astype(float)
    e_xx = np.dot(X_counter_group.T, X_counter_group)

    cross_cov = e_xx / (X_counter_group.shape[0] - 1)  # unbiased cov estimate
    return cross_cov


def cross_covariance_score(X, a, normalize=False, sum_scores=True,
                           func=np.max, off_diagonal_only=False):
    """
    Reduce the cross-covariance matrix into a single score
    Args:
        X (pd.DataFrame): Covariate matrix of size (num_subjects, num_features).
        a (pd.Series): Treatment assignment of size (num_subjects,).
        func (callable): function to apply on the cross-covariance matrix
        normalize (bool): If True, standardize features by removing the mean
                          and scaling to unit variances.
        off_diagonal_only (bool): The diagonal of the cross-covariance matrix
                                  represents variances, and by definition
                                  max{var(x),var(y)} >= cov(x,y).
                                  If True, set on zero the diagonal of the
                                  cross-covariance matrix, focusing on maximum
                                  of cov elements.
                                  otherwise, focusing on maximum of
                                  var elements.

        sum_scores (bool): If True, sum the scores of the different groups,
                           otherwise returns a list of scores.
    Returns:
        float|list: non-negative scores
    """
    treatment_values = sorted(pd.unique(a))
    if normalize:
        X = pd.DataFrame((X - X.mean(axis=0)) / X.std(axis=0), index=X.index)

    scores = list()
    for treatment in treatment_values:
        cross_cov = np.abs(cross_covariance(X, a, mean_group=treatment))
        if off_diagonal_only:
            cross_cov -= np.diag(np.diag(cross_cov))
        scores.append(func(cross_cov))
    return np.sum(scores) if sum_scores else scores



import unittest
from causallib.positivity.datasets.positivity_data_simulator import make_1d_overlap_data
from causallib.positivity.datasets.pizza_data_simulator import pizza


class TestPositivityDataSim(unittest.TestCase):

    def test_get_1d_data(self):
        treatment_bounds = (0, 75)
        control_bounds = (25, 100)
        X, a = make_1d_overlap_data(
            treatment_bounds=treatment_bounds, control_bounds=control_bounds)
        self.assertEqual(X[a == 0].values.min(), control_bounds[0])
        self.assertEqual(X[a == 0].values.max(), control_bounds[1]-1)
        self.assertEqual(X[a == 1].values.min(), treatment_bounds[0])
        self.assertEqual(X[a == 1].values.max(), treatment_bounds[1]-1)


class TestPositivityPizzaData(unittest.TestCase):
    def test_pizza_data(self):
        X, a = pizza(seed=0, n_samples=10000)
        self.assertAlmostEqual(
            X.loc[a == 0, 0].values.min(), X.loc[a == 1, 0].values.min(),
            delta=0.1,  # depends on the density of the points
        )
        self.assertAlmostEqual(
            X.loc[a == 0, 1].values.min(), X.loc[a == 1, 1].values.min(),
            delta=0.1,  # depends on the density of the points
        )
        self.assertAlmostEqual(
            X.loc[a == 0, 0].values.max(), X.loc[a == 1, 0].values.max(),
            delta=0.1,  # depends on the density of the points
        )
        self.assertAlmostEqual(
            X.loc[a == 0, 1].values.max(), X.loc[a == 1, 1].values.max(),
            delta=0.1,  # depends on the density of the points
        )

    def test_dimensions(self):
        X, a = pizza(n_dim=3, n_samples=10)
        self.assertEqual(3, X.shape[1])
        self.assertEqual(10, X.shape[0])
        self.assertEqual(10, a.shape[0])



# todo:
#   adding more datasets as positivity benchmarks
#   and maybe order it nicely in a Class of Simulated_data

import numpy as np
import pandas as pd


def _rotate_data(X, angle_rotation):
    """ rotate only the first 2d """
    angle = np.deg2rad(angle_rotation)
    rot_2d = np.array([[np.cos(angle), -np.sin(angle)],
                       [np.sin(angle), np.cos(angle)]])
    rot_mat = np.eye(X.shape[1])
    rot_mat[:2, :2] = rot_2d
    X_rotated = np.dot(X, rot_mat)
    return pd.DataFrame(X_rotated)


def _probability_of_being_treated(X, angle_slice):
    """ based on the amount of slicing in the data """
    angle = np.deg2rad(angle_slice)
    slope = np.sin(angle)/np.cos(angle)
    slicing_cond = (0 < X.iloc[:, 1]) & ((X.iloc[:, 1] / slope) < X.iloc[:, 0])
    p = 0.5 + 0.5 * slicing_cond
    return p


def pizza(n_dim=2, n_samples=1000,
          angle_rotation=45, angle_slice=90, seed=0):
    """
    Rotate the data and create slice with strict non-overlapping,
    i.e. the propensity score given this sub space is equal to 1.

    If n_dim>2, uniform distribution over the covariate space, where
    the non-overlapping area is the same as n_dim=2, meaning that the
    other covariates are completely overlapping.
    Args:
        n_dim (int): number of dimensions, have to be equal or bigger than 2.
        n_samples (int): number of samples
        angle_rotation (float): the angle of rotation in degrees
        angle_slice (float): the angle of sliced out area, the bigger the angle
                             the wider the non-overlapping area,
                             ranging from [0, 180]
        seed (None | int):

    Returns:

    """
    np.random.seed(seed)
    X = np.random.uniform(-np.ones(n_dim),
                          np.ones(n_dim),
                          size=(n_samples, n_dim)) / np.sqrt(n_dim)
    X_rotated = _rotate_data(X, angle_rotation)
    p = _probability_of_being_treated(X_rotated, angle_slice)
    a = pd.Series(np.random.binomial(1, p), name='a')
    return X_rotated, a
